from typing import Union, Tuple, Optional

from genomics_data_index.storage.model.QueryFeatureHGVS import QueryFeatureHGVS
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI


class NucleotideMutationTranslater:

    @classmethod
    def convert_deletion(cls, deletion: Union[str, int]) -> int:
        if isinstance(deletion, str):
            if deletion.isdigit():
                deletion = int(deletion)
            else:
                if not set(deletion).issubset({'A', 'T', 'C', 'G'}):
                    raise Exception('Deletion must either be an integer or a string with alphabet {A,T,C,G}'
                                    f': {deletion}')
                deletion = len(deletion)
        elif isinstance(deletion, int):
            if deletion < 0:
                raise Exception(f'ref=[{deletion}] must be a non-negative integer')
        else:
            raise Exception(f'ref=[{deletion}] must be either a string or a non-negative integer')

        return deletion

    @classmethod
    def from_spdi(cls, spdi: str, convert_deletion=True) -> Tuple[str, int, Union[int, str], str]:
        if spdi is None:
            raise Exception('Cannot parse value spdi=None')

        values = spdi.split(':')
        if len(values) != 4:
            raise Exception(f'Incorrect number of items for spdi=[{spdi}]')
        else:
            position = int(values[1])

            if convert_deletion:
                deletion = cls.convert_deletion(values[2])
            else:
                deletion = values[2]
                if deletion.isdigit():
                    deletion = int(deletion)
                elif deletion == '':
                    raise Exception(f'deletion=[{deletion}] but convert_deletion is False')

            if position < 0:
                raise Exception(f'Position must be non-negative: {position}')

            return str(values[0]), position, deletion, str(values[3])

    @classmethod
    def to_spdi(cls, sequence_name: str, position: int, ref: Union[str, int], alt: str,
                convert_deletion: bool = True) -> str:
        if position < 0:
            raise Exception(f'Position must be non-negative: {position}')

        if convert_deletion:
            ref = cls.convert_deletion(ref)

        return f'{sequence_name}:{position}:{ref}:{alt}'

    @classmethod
    def to_db_feature(cls, feature: QueryFeatureMutationSPDI) -> QueryFeatureMutationSPDI:
        new_id = cls.to_spdi(feature.scope, feature.position, feature.ref, feature.alt)
        return QueryFeatureMutationSPDI(new_id)

    @classmethod
    def to_hgvs_id(cls, sequence_name: str, gene: Optional[str], variant: str) -> Optional[str]:
        return QueryFeatureHGVS.create(sequence_name=sequence_name,
                                       gene=gene,
                                       variant=variant).id



import pandas as pd
import pytest

from genomics_data_index.storage.io.mutation.VcfSnpEffAnnotationParser import VcfSnpEffAnnotationParser, \
    InvalidSnpEffVcfError


@pytest.fixture
def vcf_snpeff_annotation_parser() -> VcfSnpEffAnnotationParser:
    return VcfSnpEffAnnotationParser()


@pytest.fixture
def mock_snpeff_infos():
    class MockAnn():
        def __init__(self):
            pass

        desc = ("Functional annotations: 'Allele | Annotation | Annotation_Impact | Gene_Name | Gene_ID"
                " | Feature_Type | Feature_ID | Transcript_BioType | Rank | HGVS.c | HGVS.p"
                " | cDNA.pos / cDNA.length | CDS.pos / CDS.length | AA.pos / AA.length | Distance"
                " | ERRORS / WARNINGS / INFO'")

    return {
        'ANN': MockAnn()
    }


@pytest.fixture
def mock_vcf_df_with_ann_single() -> pd.DataFrame:
    return pd.DataFrame([
        ['NC_011083', 140658, 'C', 'A',
         {'ANN': ('A|missense_variant|MODERATE|murF|SEHA_RS01180|transcript|SEHA_RS01180|'
                  'protein_coding|1/1|c.497C>A|p.Ala166Glu|497/1359|497/1359|166/452||')}
         ],
    ], columns=[
        'CHROM', 'POS', 'REF', 'ALT', 'INFO',
    ])


@pytest.fixture
def mock_vcf_df_without_ann_single() -> pd.DataFrame:
    return pd.DataFrame([
        ['NC_011083', 140658, 'C', 'A', {}],
    ], columns=[
        'CHROM', 'POS', 'REF', 'ALT', 'INFO',
    ])


@pytest.fixture
def mock_vcf_df_without_ann_multiple() -> pd.DataFrame:
    return pd.DataFrame([
        ['NC_011083', 140658, 'C', 'A', {}],
        ['NC_011083', 203200, 'C', 'T', {}],
    ], columns=[
        'CHROM', 'POS', 'REF', 'ALT', 'INFO',
    ])


@pytest.fixture
def mock_vcf_df_with_and_without_ann() -> pd.DataFrame:
    return pd.DataFrame([
        ['NC_011083', 140658, 'C', 'A',
         {'ANN': [('A|missense_variant|MODERATE|murF|SEHA_RS01180|transcript|SEHA_RS01180|'
                   'protein_coding|1/1|c.497C>A|p.Ala166Glu|497/1359|497/1359|166/452||'),
                  ('A|upstream_gene_variant|MODIFIER|mraY|SEHA_RS01185|transcript|SEHA_RS01185|'
                   'protein_coding||c.-856C>A|||||856|'),
                  ('A|upstream_gene_variant|MODIFIER|murD|SEHA_RS01190|transcript|SEHA_RS01190|'
                   'protein_coding||c.-1941C>A|||||1941|')]}
         ],
        ['NC_011083', 203200, 'C', 'T', {'ANN': []}],
    ], columns=[
        'CHROM', 'POS', 'REF', 'ALT', 'INFO',
    ])


@pytest.fixture
def mock_vcf_df_empty() -> pd.DataFrame:
    return pd.DataFrame(columns=[
        'CHROM', 'POS', 'REF', 'ALT', 'INFO',
    ])


@pytest.fixture
def mock_snpeff_infos_empty():
    return {}


@pytest.fixture
def mock_vcf_df_with_ann_multiple_entries_single_sample() -> pd.DataFrame:
    return pd.DataFrame([
        ['NC_011083', 140658, 'C', 'A',
         {'ANN': [('A|missense_variant|MODERATE|murF|SEHA_RS01180|transcript|SEHA_RS01180|'
                   'protein_coding|1/1|c.497C>A|p.Ala166Glu|497/1359|497/1359|166/452||'),
                  ('A|upstream_gene_variant|MODIFIER|mraY|SEHA_RS01185|transcript|SEHA_RS01185|'
                   'protein_coding||c.-856C>A|||||856|'),
                  ('A|upstream_gene_variant|MODIFIER|murD|SEHA_RS01190|transcript|SEHA_RS01190|'
                   'protein_coding||c.-1941C>A|||||1941|')]}
         ],
    ], columns=[
        'CHROM', 'POS', 'REF', 'ALT', 'INFO',
    ])


@pytest.fixture
def mock_vcf_df_single_sample_annotations() -> pd.DataFrame:
    return pd.DataFrame([
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'mraY', 'SEHA_RS01185', 'transcript', 'protein_coding',
         'c.-856C>A', pd.NA, pd.NA],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
         'c.497C>A', 'p.Ala166Glu', '497/1359'],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'murD', 'SEHA_RS01190', 'transcript', 'protein_coding',
         'c.-1941C>A', pd.NA, pd.NA]
    ], columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p',
                'ANN.cDNA.pos / cDNA.length',
                ])


@pytest.fixture
def mock_vcf_df_single_sample_annotations_some_na_values() -> pd.DataFrame:
    return pd.DataFrame([
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'mraY', 'SEHA_RS01185', 'transcript', 'protein_coding',
         'c.-856C>A', pd.NA, pd.NA],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', pd.NA,
         'c.497C>A', 'p.Ala166Glu', '497/1359'],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'murD', 'SEHA_RS01190', 'transcript', 'protein_coding',
         'c.-1941C>A', pd.NA, pd.NA]
    ], columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p',
                'ANN.cDNA.pos / cDNA.length',
                ])


@pytest.fixture
def mock_vcf_df_multiple_sample_annotations() -> pd.DataFrame:
    return pd.DataFrame([
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'mraY', 'SEHA_RS01185', 'transcript', 'protein_coding',
         'c.-856C>A', pd.NA, pd.NA],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
         'c.497C>A', 'p.Ala166Glu', '497/1359'],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'murD', 'SEHA_RS01190', 'transcript', 'protein_coding',
         'c.-1941C>A', pd.NA, pd.NA],
        ['SampleA', 'NC_011083', 203200, 'C', 'T', 'snp', 'SampleA.vcf', 'NC_011083:203200:C:T',
         'T', 'missense_variant', 'MODERATE', 'SEHA_RS01460', 'SEHA_RS01460', 'transcript',
         'protein_coding', 'c.602C>T', 'p.Thr201Met', '602/927'],
        ['SampleA', 'NC_011083', 203200, 'C', 'T', 'snp', 'SampleA.vcf', 'NC_011083:203200:C:T',
         'T', 'upstream_gene_variant', 'MODIFIER', 'SEHA_RS01455', 'SEHA_RS01455',
         'transcript', 'protein_coding', 'c.-2172G>A', pd.NA, pd.NA],
        ['SampleA', 'NC_011083', 203200, 'C', 'T', 'snp', 'SampleA.vcf', 'NC_011083:203200:C:T',
         'T', 'upstream_gene_variant', 'MODIFIER', 'SEHA_RS01455', 'SEHA_RS01455',
         'transcript', 'protein_coding', 'c.-710G>A', pd.NA, pd.NA],
    ], columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p',
                'ANN.cDNA.pos / cDNA.length',
                ])


@pytest.fixture
def mock_vcf_df_single_sample_annotations_sars_cov_2() -> pd.DataFrame:
    return pd.DataFrame([
        ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
         'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
         'c.3683A>G', 'p.D1228G', '3683/21291'],
        ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
         'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
         'c.3683A>G', 'p.D1228G', '3683/13218'],
        ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
         'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
         'c.1229A>G', 'p.D410G', '1229/5835'],
        ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
         'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
         'c.1229A>G', 'p.D410G', '1229/5835'],
    ], columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p',
                'ANN.cDNA.pos / cDNA.length',
                ])


@pytest.fixture
def mock_vcf_df_single_sample_annotations_rv_sars_cov_2() -> pd.DataFrame:
    return pd.DataFrame([
        ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
         'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
         'c.1229A>G', 'p.D410G', '1229/5835'],
        ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
         'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
         'c.1229A>G', 'p.D410G', '1229/5835'],
        ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
         'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
         'c.3683A>G', 'p.D1228G', '3683/13218'],
        ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
         'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
         'c.3683A>G', 'p.D1228G', '3683/21291'],
    ], columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p',
                'ANN.cDNA.pos / cDNA.length',
                ])


@pytest.fixture
def mock_vcf_df_multiple_sample_one_empty() -> pd.DataFrame:
    return pd.DataFrame([
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'mraY', 'SEHA_RS01185', 'transcript', 'protein_coding',
         'c.-856C>A', pd.NA, pd.NA],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
         'c.497C>A', 'p.Ala166Glu', '497/1359'],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'murD', 'SEHA_RS01190', 'transcript', 'protein_coding',
         'c.-1941C>A', pd.NA, pd.NA],
        ['SampleA', 'NC_011083', 203200, 'C', 'T', 'snp', 'SampleA.vcf', 'NC_011083:203200:C:T'] + [pd.NA] * 10,
    ], columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p',
                'ANN.cDNA.pos / cDNA.length',
                ])


@pytest.fixture
def mock_vcf_df_multiple_sample_one_invalid() -> pd.DataFrame:
    return pd.DataFrame([
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'mraY', 'SEHA_RS01185', 'transcript', 'protein_coding',
         'c.-856C>A', pd.NA, pd.NA],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
         'c.497C>A', 'p.Ala166Glu', '497/1359'],
        ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
         'A', 'upstream_gene_variant', 'MODIFIER', 'murD', 'SEHA_RS01190', 'transcript', 'protein_coding',
         'c.-1941C>A', pd.NA, pd.NA],

        # This one is invalid because snpeff 'Allele' refers to some other variant (indicating it's a compound variant)
        # I want to ignore all compound variants because it would be very difficult to index them in my software.
        ['SampleA', 'NC_011083', 203200, 'C', 'T', 'snp', 'SampleA.vcf', 'NC_011083:203200:C:T',
         'T:123456_A>T', 'missense_variant', 'MODERATE', 'SEHA_RS01460', 'SEHA_RS01460', 'transcript',
         'protein_coding', 'c.602C>T', 'p.Thr201Met', '602/927'],
    ], columns=['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p',
                'ANN.cDNA.pos / cDNA.length'
                ])


@pytest.fixture
def mock_vcf_df_with_ann_multiple_entries_multiple_samples() -> pd.DataFrame:
    return pd.DataFrame([
        ['NC_011083', 140658, 'C', 'A',
         {'ANN': [('A|missense_variant|MODERATE|murF|SEHA_RS01180|transcript|SEHA_RS01180|'
                   'protein_coding|1/1|c.497C>A|p.Ala166Glu|497/1359|497/1359|166/452||'),
                  ('A|upstream_gene_variant|MODIFIER|mraY|SEHA_RS01185|transcript|SEHA_RS01185|'
                   'protein_coding||c.-856C>A|||||856|'),
                  ('A|upstream_gene_variant|MODIFIER|murD|SEHA_RS01190|transcript|SEHA_RS01190|'
                   'protein_coding||c.-1941C>A|||||1941|')]}
         ],
        ['NC_011083', 203200, 'C', 'T',
         {'ANN': [('T|missense_variant|MODERATE|SEHA_RS01460|SEHA_RS01460|transcript|SEHA_RS01460|'
                   'protein_coding|1/1|c.602C>T|p.Thr201Met|602/927|602/927|201/308||'),
                  ('T|upstream_gene_variant|MODIFIER|SEHA_RS01445|SEHA_RS01445|transcript|SEHA_RS01445|'
                   'protein_coding||c.-2172G>A|||||2172|'),
                  ('T|upstream_gene_variant|MODIFIER|can|SEHA_RS01455|transcript|SEHA_RS01455|'
                   'protein_coding||c.-710G>A|||||710|')]}
         ],
    ], columns=[
        'CHROM', 'POS', 'REF', 'ALT', 'INFO',
    ])


@pytest.fixture
def mock_snpeff_infos_invalid():
    class MockAnn():
        def __init__(self):
            pass

        desc = 'invalid'

    return {
        'ANN': MockAnn()
    }


def test_parse_annotation_headers(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser, mock_snpeff_infos):
    headers_list = vcf_snpeff_annotation_parser.parse_annotation_headers(mock_snpeff_infos)

    assert ['Allele', 'Annotation', 'Annotation_Impact', 'Gene_Name', 'Gene_ID', 'Feature_Type', 'Feature_ID',
            'Transcript_BioType', 'Rank', 'HGVS.c', 'HGVS.p', 'cDNA.pos / cDNA.length', 'CDS.pos / CDS.length',
            'AA.pos / AA.length', 'Distance', 'ERRORS / WARNINGS / INFO'] == headers_list


def test_parse_annotation_headers_invalid(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                          mock_snpeff_infos_invalid):
    with pytest.raises(InvalidSnpEffVcfError) as execinfo:
        vcf_snpeff_annotation_parser.parse_annotation_headers(mock_snpeff_infos_invalid)
    assert "Found 'ANN' in VCF information but description" in str(execinfo.value)


def test_parse_annotation_headers_no_annotation(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                mock_snpeff_infos_empty):
    headers_list = vcf_snpeff_annotation_parser.parse_annotation_headers(mock_snpeff_infos_empty)

    assert [] == headers_list


def test_parse_annotation_entries_single(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                         mock_snpeff_infos, mock_vcf_df_with_ann_single: pd.DataFrame):
    headers_list = vcf_snpeff_annotation_parser.parse_annotation_headers(mock_snpeff_infos)
    ann_entries_df = vcf_snpeff_annotation_parser.parse_annotation_entries(vcf_ann_headers=headers_list,
                                                                           vcf_df=mock_vcf_df_with_ann_single)

    assert ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p', 'ANN.cDNA.pos / cDNA.length',
            'VARIANT_ID'] == list(
        ann_entries_df.columns)
    assert 1 == len(ann_entries_df)
    assert [0] == list(ann_entries_df.index)
    assert ['A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu', '497/1359', 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[0])


def test_parse_annotation_entries_multiple_entries_single_sample(
        vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
        mock_snpeff_infos, mock_vcf_df_with_ann_multiple_entries_single_sample: pd.DataFrame):
    headers_list = vcf_snpeff_annotation_parser.parse_annotation_headers(mock_snpeff_infos)
    ann_entries_df = vcf_snpeff_annotation_parser.parse_annotation_entries(vcf_ann_headers=headers_list,
                                                                           vcf_df=mock_vcf_df_with_ann_multiple_entries_single_sample)

    assert ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p', 'ANN.cDNA.pos / cDNA.length',
            'VARIANT_ID'] == list(
        ann_entries_df.columns)
    assert 3 == len(ann_entries_df)
    ann_entries_df = ann_entries_df.sort_values(['original_index', 'ANN.Gene_ID'])
    assert [0, 0, 0] == list(ann_entries_df.index)
    assert ['A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu', '497/1359', 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[0])
    assert ['A', 'upstream_gene_variant', 'MODIFIER', 'mraY', 'SEHA_RS01185', 'transcript', 'protein_coding',
            'c.-856C>A', pd.NA, pd.NA, 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[1])
    assert ['A', 'upstream_gene_variant', 'MODIFIER', 'murD', 'SEHA_RS01190', 'transcript', 'protein_coding',
            'c.-1941C>A', pd.NA, pd.NA, 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[2])


def test_parse_annotation_entries_multiple_entries_multiple_samples(
        vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
        mock_snpeff_infos, mock_vcf_df_with_ann_multiple_entries_multiple_samples: pd.DataFrame):
    headers_list = vcf_snpeff_annotation_parser.parse_annotation_headers(mock_snpeff_infos)
    ann_entries_df = vcf_snpeff_annotation_parser.parse_annotation_entries(vcf_ann_headers=headers_list,
                                                                           vcf_df=mock_vcf_df_with_ann_multiple_entries_multiple_samples)

    assert ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p', 'ANN.cDNA.pos / cDNA.length',
            'VARIANT_ID'] == list(
        ann_entries_df.columns)
    assert 6 == len(ann_entries_df)
    ann_entries_df = ann_entries_df.sort_values(['original_index', 'ANN.Gene_ID'])
    assert [0, 0, 0, 1, 1, 1] == list(ann_entries_df.index)
    assert ['A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu', '497/1359', 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[0])
    assert ['A', 'upstream_gene_variant', 'MODIFIER', 'mraY', 'SEHA_RS01185', 'transcript', 'protein_coding',
            'c.-856C>A', pd.NA, pd.NA, 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[1])
    assert ['A', 'upstream_gene_variant', 'MODIFIER', 'murD', 'SEHA_RS01190', 'transcript', 'protein_coding',
            'c.-1941C>A', pd.NA, pd.NA, 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[2])

    assert ['T', 'upstream_gene_variant', 'MODIFIER', 'SEHA_RS01445', 'SEHA_RS01445', 'transcript', 'protein_coding',
            'c.-2172G>A', pd.NA, pd.NA, 'NC_011083:203200:C:T'] == list(ann_entries_df.iloc[3])
    assert ['T', 'upstream_gene_variant', 'MODIFIER', 'can', 'SEHA_RS01455', 'transcript', 'protein_coding',
            'c.-710G>A', pd.NA, pd.NA, 'NC_011083:203200:C:T'] == list(ann_entries_df.iloc[4])
    assert ['T', 'missense_variant', 'MODERATE', 'SEHA_RS01460', 'SEHA_RS01460', 'transcript', 'protein_coding',
            'c.602C>T', 'p.Thr201Met', '602/927', 'NC_011083:203200:C:T'] == list(ann_entries_df.iloc[5])


def test_parse_annotation_entries_no_annotation_single_sample(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                              mock_vcf_df_without_ann_single: pd.DataFrame):
    ann_entries_df = vcf_snpeff_annotation_parser.parse_annotation_entries(vcf_ann_headers=[],
                                                                           vcf_df=mock_vcf_df_without_ann_single)

    assert ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p', 'ANN.cDNA.pos / cDNA.length',
            'VARIANT_ID'] == list(
        ann_entries_df.columns)
    assert 1 == len(ann_entries_df)
    assert [0] == list(ann_entries_df.index)
    assert {True} == set(ann_entries_df.drop('VARIANT_ID', axis='columns').iloc[0].isna())
    assert 'NC_011083:140658:C:A' == ann_entries_df['VARIANT_ID'].iloc[0]


def test_parse_annotation_entries_no_annotation_multiple_sample(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                                mock_vcf_df_without_ann_multiple: pd.DataFrame):
    ann_entries_df = vcf_snpeff_annotation_parser.parse_annotation_entries(vcf_ann_headers=[],
                                                                           vcf_df=mock_vcf_df_without_ann_multiple)

    assert ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p', 'ANN.cDNA.pos / cDNA.length',
            'VARIANT_ID'] == list(
        ann_entries_df.columns)
    assert 2 == len(ann_entries_df)
    assert [0, 1] == list(ann_entries_df.index)
    assert {True} == set(ann_entries_df.drop('VARIANT_ID', axis='columns').iloc[0].isna())
    assert 'NC_011083:140658:C:A' == ann_entries_df['VARIANT_ID'].iloc[0]
    assert {True} == set(ann_entries_df.drop('VARIANT_ID', axis='columns').iloc[1].isna())
    assert 'NC_011083:203200:C:T' == ann_entries_df['VARIANT_ID'].iloc[1]


def test_parse_annotation_entries_some_with_some_without_annotations(
        vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
        mock_snpeff_infos,
        mock_vcf_df_with_and_without_ann: pd.DataFrame):
    headers_list = vcf_snpeff_annotation_parser.parse_annotation_headers(mock_snpeff_infos)
    ann_entries_df = vcf_snpeff_annotation_parser.parse_annotation_entries(vcf_ann_headers=headers_list,
                                                                           vcf_df=mock_vcf_df_with_and_without_ann)

    assert ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p', 'ANN.cDNA.pos / cDNA.length',
            'VARIANT_ID'] == list(
        ann_entries_df.columns)
    assert 4 == len(ann_entries_df)
    ann_entries_df = ann_entries_df.sort_values(['original_index', 'ANN.Gene_ID'])
    assert [0, 0, 0, 1] == list(ann_entries_df.index)
    assert ['A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu', '497/1359', 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[0])
    assert ['A', 'upstream_gene_variant', 'MODIFIER', 'mraY', 'SEHA_RS01185', 'transcript', 'protein_coding',
            'c.-856C>A', pd.NA, pd.NA, 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[1])
    assert ['A', 'upstream_gene_variant', 'MODIFIER', 'murD', 'SEHA_RS01190', 'transcript', 'protein_coding',
            'c.-1941C>A', pd.NA, pd.NA, 'NC_011083:140658:C:A'] == list(ann_entries_df.iloc[2])

    assert {True} == set(ann_entries_df.drop('VARIANT_ID', axis='columns').iloc[3].isna())
    assert 'NC_011083:203200:C:T' == ann_entries_df['VARIANT_ID'].iloc[3]


def test_parse_annotation_entries_empty(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                        mock_vcf_df_empty: pd.DataFrame):
    ann_entries_df = vcf_snpeff_annotation_parser.parse_annotation_entries(vcf_ann_headers=[],
                                                                           vcf_df=mock_vcf_df_empty)

    assert ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p', 'ANN.cDNA.pos / cDNA.length',
            'VARIANT_ID'] == list(
        ann_entries_df.columns)
    assert 0 == len(ann_entries_df)


def test_select_variant_annotations_single_sample(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                  mock_vcf_df_single_sample_annotations: pd.DataFrame):
    ann_entries_df = vcf_snpeff_annotation_parser.select_variant_annotations(mock_vcf_df_single_sample_annotations)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        ann_entries_df.columns)
    assert 1 == len(ann_entries_df)
    assert ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
            'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu'] == list(ann_entries_df.iloc[0])


def test_select_variant_annotations_some_na_values(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                   mock_vcf_df_single_sample_annotations_some_na_values: pd.DataFrame):
    ann_entries_df = vcf_snpeff_annotation_parser.select_variant_annotations(
        mock_vcf_df_single_sample_annotations_some_na_values)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        ann_entries_df.columns)
    assert 1 == len(ann_entries_df)
    assert ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
            'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'NA',
            'c.497C>A', 'p.Ala166Glu'] == list(ann_entries_df.iloc[0].fillna('NA'))


def test_select_variant_annotations_single_sample_sars_cov_2(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                             mock_vcf_df_single_sample_annotations_sars_cov_2: pd.DataFrame):
    ann_entries_df = vcf_snpeff_annotation_parser.select_variant_annotations(
        mock_vcf_df_single_sample_annotations_sars_cov_2)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        ann_entries_df.columns)
    assert 1 == len(ann_entries_df)
    assert ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
            'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
            'c.3683A>G', 'p.D1228G'] == list(ann_entries_df.iloc[0])


def test_select_variant_annotations_rv_single_sample_sars_cov_2(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                                mock_vcf_df_single_sample_annotations_rv_sars_cov_2: pd.DataFrame):
    ann_entries_df = vcf_snpeff_annotation_parser.select_variant_annotations(
        mock_vcf_df_single_sample_annotations_rv_sars_cov_2)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        ann_entries_df.columns)
    assert 1 == len(ann_entries_df)
    assert ['SampleA', 'NC_045512.2', 3948, 'A', 'G', 'SNP', 'SampleA.vcf', 'NC_045512.2:3948:A:G',
            'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript', 'protein_coding',
            'c.3683A>G', 'p.D1228G'] == list(ann_entries_df.iloc[0])


def test_select_variant_annotations_multiple_sample(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                    mock_vcf_df_multiple_sample_annotations: pd.DataFrame):
    ann_entries_df = vcf_snpeff_annotation_parser.select_variant_annotations(mock_vcf_df_multiple_sample_annotations)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        ann_entries_df.columns)
    assert 2 == len(ann_entries_df)
    assert ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
            'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu'] == list(ann_entries_df.iloc[0])
    assert ['SampleA', 'NC_011083', 203200, 'C', 'T', 'snp', 'SampleA.vcf', 'NC_011083:203200:C:T',
            'T', 'missense_variant', 'MODERATE', 'SEHA_RS01460', 'SEHA_RS01460', 'transcript', 'protein_coding',
            'c.602C>T', 'p.Thr201Met'] == list(ann_entries_df.iloc[1])


def test_select_variant_annotations_one_no_annotation(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                      mock_vcf_df_multiple_sample_one_empty: pd.DataFrame):
    annotation_columns = ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                          'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    ann_entries_df = vcf_snpeff_annotation_parser.select_variant_annotations(mock_vcf_df_multiple_sample_one_empty)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        ann_entries_df.columns)
    assert 2 == len(ann_entries_df)
    assert ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
            'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu'] == list(ann_entries_df.iloc[0])
    assert ['SampleA', 'NC_011083', 203200, 'C', 'T', 'snp', 'SampleA.vcf',
            'NC_011083:203200:C:T'] == list(ann_entries_df.drop(annotation_columns, axis='columns').iloc[1])
    assert {True} == set(ann_entries_df[annotation_columns].iloc[1].isna().tolist())


def test_select_variant_annotations_one_invalid_annotation(vcf_snpeff_annotation_parser: VcfSnpEffAnnotationParser,
                                                           mock_vcf_df_multiple_sample_one_invalid: pd.DataFrame):
    annotation_columns = ['ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                          'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    ann_entries_df = vcf_snpeff_annotation_parser.select_variant_annotations(mock_vcf_df_multiple_sample_one_invalid)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        ann_entries_df.columns)
    assert 2 == len(ann_entries_df)
    assert ['SampleA', 'NC_011083', 140658, 'C', 'A', 'snp', 'SampleA.vcf', 'NC_011083:140658:C:A',
            'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu'] == list(ann_entries_df.iloc[0])
    assert ['SampleA', 'NC_011083', 203200, 'C', 'T', 'snp', 'SampleA.vcf',
            'NC_011083:203200:C:T'] == list(ann_entries_df.drop(annotation_columns, axis='columns').iloc[1])
    assert {True} == set(ann_entries_df[annotation_columns].iloc[1].isna().tolist())



import shutil
import warnings
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

warnings.filterwarnings("ignore", category=DeprecationWarning)

from genomics_data_index.test.integration import test_project_dir
from genomics_data_index.configuration.Project import Project
from genomics_data_index.configuration.ConfigFile import ConfigFile


def test_create_new_project():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)

        Project.initialize_project(tmp_dir / 'project')

        assert (tmp_dir / 'project').exists()
        assert (tmp_dir / 'project' / '.gdi-data').exists()
        assert (tmp_dir / 'project' / 'gdi-config.yaml').exists()

        config = ConfigFile.read_config(tmp_dir / 'project' / 'gdi-config.yaml')
        assert config.database_connection is None
        assert config.database_dir == Path('.gdi-data')
        assert config.sqlite_database == Path('.gdi-data') / 'gdi-db.sqlite'


def test_from_existing_project():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        project_dir = tmp_dir / 'project'
        shutil.copytree(test_project_dir, project_dir)

        project = Project(project_dir)

        assert f'sqlite:///{project_dir}/.gdi-data/gdi-db.sqlite' == project.get_database_connection_str()
        assert project_dir / '.gdi-data' == project.get_database_dir()

        connection = project.create_connection()
        assert connection.filesystem_storage.variation_dir.parent == project_dir / '.gdi-data'


def test_initialize_from_existing_project():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        project_dir = tmp_dir / 'project'
        shutil.copytree(test_project_dir, project_dir)

        with pytest.raises(Exception) as execinfo:
            Project.initialize_project(project_dir)
        assert 'has already been initialized' in str(execinfo.value)



import abc
from typing import Generator, List

from genomics_data_index.storage.io.SampleData import SampleData


class SampleFilesProcessor(abc.ABC):

    def __init__(self):
        pass

    @abc.abstractmethod
    def process(self, sample_data: List[SampleData]) -> Generator[SampleData, None, None]:
        pass



import abc
import logging
from typing import Union, Iterable, Dict, Any, Optional

from ete3 import RectFace, Face, CircleFace

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.viewer.TreeSamplesVisual import TreeSamplesVisual

logger = logging.getLogger(__name__)


class AbstractAnnotationTreeSamplesVisual(TreeSamplesVisual, abc.ABC):
    """
    A tree style for adding annotations (columns next to leaves of a tree).
    """

    ANNOTATE_KINDS = TreeSamplesVisual.LEGEND_FACE_KINDS

    def __init__(self, samples: Union[SamplesQuery, Iterable[str]],
                 legend_nodesize: int,
                 legend_fontsize: int,
                 legend_columns: Dict[str, int],
                 annotate_kind: str,
                 annotate_border_width: Optional[int],
                 annotate_margin: Optional[int],
                 ):
        super().__init__(samples=samples,
                         legend_nodesize=legend_nodesize,
                         legend_fontsize=legend_fontsize,
                         legend_columns=legend_columns)

        if annotate_kind not in self.ANNOTATE_KINDS:
            raise Exception(f'annotate_kind=[{annotate_kind}] is invalid. Must be one of {self.ANNOTATE_KINDS}')
        else:
            self._annotate_kind = annotate_kind

        self._annotate_border_width = annotate_border_width
        self._annotate_margin = annotate_margin

    def _build_annotate_face(self, width: int, height: int, border_color: str, bgcolor: str,
                             opacity: float, label: Union[str, Dict[str, Any]] = None) -> Face:
        if self._annotate_kind == 'rect' or self._annotate_kind == 'rectangle' or self._annotate_kind == 'r':
            rf = RectFace(width=width, height=height, fgcolor=None, bgcolor=bgcolor, label=label)
            rf.border.width = self._annotate_border_width
            rf.margin_top = self._annotate_margin
            rf.margin_bottom = self._annotate_margin
            rf.margin_left = self._annotate_margin
            rf.margin_right = self._annotate_margin
            rf.border.color = border_color
            rf.background.color = bgcolor
            rf.opacity = opacity
            rf.hz_align = 1
            rf.vt_align = 1
            return rf
        elif self._annotate_kind == 'circle' or self._annotate_kind == 'circ' or self._annotate_kind == 'c':
            # Make circle radius such that it fits in bounding box defined by width and height
            min_dimension = min(width, height)
            radius = min_dimension / 2

            cf = CircleFace(radius=radius, color=bgcolor, label=label)
            cf.border.width = self._annotate_border_width
            cf.margin_top = self._annotate_margin
            cf.margin_bottom = self._annotate_margin
            cf.margin_left = self._annotate_margin
            cf.margin_right = self._annotate_margin
            cf.border.color = border_color
            cf.opacity = opacity
            cf.hz_align = 1
            cf.vt_align = 1
            return cf
        else:
            raise Exception(f'Invalid value for annotate_kind={self._annotate_kind}.'
                            f' Must be one of {self.LEGEND_FACE_KINDS}')



import logging
import multiprocessing as mp
from typing import List, Generator

import pandas as pd

from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData
from genomics_data_index.storage.io.mutation.variants_processor.VcfVariantsTableProcessor import \
    VcfVariantsTableProcessor, VcfVariantsTableProcessorFactory

logger = logging.getLogger(__name__)


class MultipleProcessVcfVariantsTableProcessor(VcfVariantsTableProcessor):

    def __init__(self, include_masked_regions: bool, ncores: int, max_chunk_size: int):
        super().__init__(include_masked_regions)
        self._ncores = ncores
        self._max_chunk_size = max_chunk_size

    def _get_chunk_size(self, number_samples: int):
        chunk_size = max(1, int(number_samples / self._ncores))
        return min(self._max_chunk_size, chunk_size)

    def _read_sample_data(self, sample_data: NucleotideSampleData) -> pd.DataFrame:
        return sample_data.read_sample_data_features(include_masked_regions=self.include_masked_regions)

    def process(self, sample_data: List[NucleotideSampleData]) -> Generator[pd.DataFrame, None, None]:
        number_samples = len(sample_data)
        chunk_size = self._get_chunk_size(number_samples)
        logger.debug(f'Starting to read features for {number_samples} samples '
                     f'with {self._ncores} cores and chunk size {chunk_size}')

        with mp.Pool(self._ncores) as pool:
            frame_vcf_masks = pool.imap_unordered(self._read_sample_data,
                                                  sample_data,
                                                  chunk_size)
            for frame_vcf_mask in frame_vcf_masks:
                yield frame_vcf_mask

        logger.debug(f'Finished reading features table from {number_samples} samples')


class MultipleProcessVcfVariantsTableProcessorFactory(VcfVariantsTableProcessorFactory):

    def __init__(self, ncores: int = 1, max_chunk_size: int = 500):
        super().__init__()
        self._ncores = ncores
        self._max_chunk_size = max_chunk_size

    def create_variants_processor(self, include_masked_regions: bool) -> VcfVariantsTableProcessor:
        return MultipleProcessVcfVariantsTableProcessor(include_masked_regions,
                                                        ncores=self._ncores,
                                                        max_chunk_size=self._max_chunk_size)



from distutils.core import setup

from setuptools import find_packages

classifiers = """
Development Status :: 4 - Beta
Environment :: Console
License :: OSI Approved :: Apache Software License
Intended Audience :: Science/Research
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Bio-Informatics
Programming Language :: Python :: 3.8
Operating System :: POSIX :: Linux
""".strip().split('\n')

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='genomics-data-index',
      version='0.6.0',
      description='Indexes genomics data (mutations, kmers, MLST) for fast querying of features.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Aaron Petkau',
      author_email='aaron.petkau@gmail.com',
      url='https://github.com/apetkau/genomics-data-index',
      license='Apache v2.0',
      classifiers=classifiers,
      install_requires=[
          # pyvcf uses the option use_2to3 which is not compatible with setuptools>=58
          'setuptools<58',

          'biopython>=1.70',
          'pandas>=1.0.0',
          'numpy',
          'scikit-bio',
          'scipy',
          'pyvcf',
          'sqlalchemy',
          'pymysql',
          'requests',

          # ete3 requires PyQt5 for some functionality <https://github.com/etetoolkit/ete/issues/354>
          'ete3',
          'PyQt5',

          # ga4gh cannot use recent versions of jsonschema
          'jsonschema==3.2.0',
          'ga4gh.vrs[extras]==0.6.2',
          'biocommons.seqrepo',
          'click',
          'click-config-file',
          'coloredlogs',
          'pyyaml',
          'pyroaring',
          'sourmash',
          'pybedtools',
          'snakemake',
          'Jinja2',
          'pathvalidate',
          'pytest',

          # I do not know exactly why these two statements are required and aren't 
          # picked up by the dependency resolver, but they seem to be necessary now.
          # 'tomli' required by 'black' and the wrong version is being installed. This fixes it.
          'tomli<2.0.0,>=0.2.6',
          'zipp',
      ],
      python_requires="<3.9",
      packages=find_packages(),
      include_package_data=True,
      entry_points={
          'console_scripts': ['gdi=genomics_data_index.cli.gdi:main'],
      },
      )



import abc
from typing import List, Set, Optional

import pandas as pd

from genomics_data_index.storage.io.SampleData import SampleData


class FeaturesReader(abc.ABC):

    def __init__(self):
        pass

    def _check_features_table_columns(self, features_df: pd.DataFrame) -> None:
        expected_columns = self._minimal_expected_columns()
        actual_columns = set(features_df.columns.tolist())
        if not expected_columns.issubset(actual_columns):
            raise Exception('Variants table does not contain expected set of columns. '
                            f'Expected {expected_columns}, actual {actual_columns}')

    @abc.abstractmethod
    def _minimal_expected_columns(self) -> Set[str]:
        pass

    def get_features_table(self) -> pd.DataFrame:
        features_df = self._read_features_table()
        self._check_features_table_columns(features_df)
        return features_df

    @abc.abstractmethod
    def get_sample_files(self, sample_name: str) -> Optional[SampleData]:
        pass

    @abc.abstractmethod
    def get_or_create_feature_file(self, sample_name: str):
        """
        Gets a file of the features associated with the sample (or creates such a file if it doesn't exist).
        :param sample_name: The sample to get features for.
        :return: A Path to the file containing all the features for the sample.
        """
        pass

    @abc.abstractmethod
    def samples_list(self) -> List[str]:
        """
        Gets a list of sample names that will be read by this reader.
        :return: A list of sample names that will be read by this reader.
        """
        pass

    def samples_set(self) -> Set[str]:
        """
        Gets a set of sample names that will be read by this reader.
        :return: A set of sample names that will be read by this reader.
        """
        return set(self.samples_list())

    @abc.abstractmethod
    def _read_features_table(self) -> pd.DataFrame:
        pass



import tempfile
from pathlib import Path

import pandas as pd
import pytest

from genomics_data_index.pipelines.SnakemakePipelineExecutor import SnakemakePipelineExecutor


def test_create_input_sample_files_assemblies_only():
    files = [Path('SampleA.fasta'),
             Path('SampleB.fasta'),
             Path('SampleC_xyz.fa'),
             Path('SampleD.something.fna.gz'),
             Path('/tmp', 'SampleE.fasta.gz')]

    executor = SnakemakePipelineExecutor()

    df = executor.create_input_sample_files(files)
    assert 5 == len(df)
    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()

    df = df.set_index('Sample', drop=False)
    assert ['SampleA', Path('SampleA.fasta'), pd.NA, pd.NA] == df.loc['SampleA'].tolist()
    assert ['SampleB', Path('SampleB.fasta'), pd.NA, pd.NA] == df.loc['SampleB'].tolist()
    assert ['SampleC_xyz', Path('SampleC_xyz.fa'), pd.NA, pd.NA] == df.loc['SampleC_xyz'].tolist()
    assert ['SampleD.something', Path('SampleD.something.fna.gz'), pd.NA, pd.NA] == df.loc['SampleD.something'].tolist()
    assert ['SampleE', Path('/tmp/SampleE.fasta.gz'), pd.NA, pd.NA] == df.loc['SampleE'].tolist()


def test_create_input_sample_files_reads_only():
    files = [Path('SampleA.fastq'),
             Path('SampleB.fastq.gz'),
             Path('SampleC_1.fastq.gz'),
             Path('SampleC_2.fastq.gz'),
             Path('/tmp', 'SampleD_xyz_R1_001.fastq'),
             Path('/tmp', 'SampleD_xyz_R2_001.fastq')]

    executor = SnakemakePipelineExecutor()

    df = executor.create_input_sample_files(files)
    assert 4 == len(df)
    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()

    df = df.set_index('Sample', drop=False)
    assert ['SampleA', pd.NA, Path('SampleA.fastq'), pd.NA] == df.loc['SampleA'].tolist()
    assert ['SampleB', pd.NA, Path('SampleB.fastq.gz'), pd.NA] == df.loc['SampleB'].tolist()
    assert ['SampleC', pd.NA, Path('SampleC_1.fastq.gz'), Path('SampleC_2.fastq.gz')] == df.loc['SampleC'].tolist()
    assert ['SampleD_xyz', pd.NA, Path('/tmp/SampleD_xyz_R1_001.fastq'),
            Path('/tmp/SampleD_xyz_R2_001.fastq')] == df.loc['SampleD_xyz'].tolist()


def test_create_input_sample_files_reads_and_assemblies():
    files = [Path('SampleA.fastq'),
             Path('SampleB.fastq.gz'),
             Path('SampleC_1.fastq.gz'),
             Path('SampleC_2.fastq.gz'),
             Path('/tmp', 'SampleD_xyz_R1_001.fastq'),
             Path('/tmp', 'SampleD_xyz_R2_001.fastq'),
             Path('..', 'SampleE_xyz.fasta'),
             Path('SampleF.fna.gz')]

    executor = SnakemakePipelineExecutor()

    df = executor.create_input_sample_files(files)
    assert 6 == len(df)
    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()

    df = df.set_index('Sample', drop=False)
    assert ['SampleA', pd.NA, Path('SampleA.fastq'), pd.NA] == df.loc['SampleA'].tolist()
    assert ['SampleB', pd.NA, Path('SampleB.fastq.gz'), pd.NA] == df.loc['SampleB'].tolist()
    assert ['SampleC', pd.NA, Path('SampleC_1.fastq.gz'), Path('SampleC_2.fastq.gz')] == df.loc['SampleC'].tolist()
    assert ['SampleD_xyz', pd.NA, Path('/tmp/SampleD_xyz_R1_001.fastq'),
            Path('/tmp/SampleD_xyz_R2_001.fastq')] == df.loc['SampleD_xyz'].tolist()
    assert ['SampleE_xyz', Path('../SampleE_xyz.fasta'), pd.NA, pd.NA] == df.loc['SampleE_xyz'].tolist()
    assert ['SampleF', Path('SampleF.fna.gz'), pd.NA, pd.NA] == df.loc['SampleF'].tolist()


def test_create_input_sample_files_reads_and_assemblies_duplicate_sample_names():
    files = [Path('SampleA.fasta'),
             Path('SampleB.fasta'),
             Path('SampleA_1.fastq.gz'),
             Path('SampleA_2.fastq.gz')]

    executor = SnakemakePipelineExecutor()

    with pytest.raises(Exception) as execinfo:
        executor.create_input_sample_files(files)

    assert 'Duplicate sample with name [SampleA]' in str(execinfo.value)


def test_create_input_sample_files_reads_and_assemblies_duplicate_sample_names_2():
    files = [Path('/tmp', 'SampleA_1.fastq.gz'),
             Path('/tmp', 'SampleA_2.fastq.gz'),
             Path('SampleA.fasta'),
             Path('SampleB.fasta')]

    executor = SnakemakePipelineExecutor()

    with pytest.raises(Exception) as execinfo:
        executor.create_input_sample_files(files)

    assert 'Duplicate sample with name [SampleA]' in str(execinfo.value)


def test_skip_samples_from_input_files():
    input_samples = pd.DataFrame([
        ['A', Path('file.fasta'), pd.NA, pd.NA],
        ['B', Path('file2.fasta'), pd.NA, pd.NA],
        ['C', pd.NA, Path('file_1.fastq'), Path('file_2.fastq')]
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    df = executor.skip_samples_from_input_files(input_samples, skip_samples=None)
    df = df.fillna('NA')
    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
    assert 3 == len(df)
    assert ['A', Path('file.fasta'), 'NA', 'NA'] == df.iloc[0].tolist()
    assert ['B', Path('file2.fasta'), 'NA', 'NA'] == df.iloc[1].tolist()
    assert ['C', 'NA', Path('file_1.fastq'), Path('file_2.fastq')] == df.iloc[2].tolist()

    df = executor.skip_samples_from_input_files(input_samples, skip_samples=set())
    df = df.fillna('NA')
    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
    assert 3 == len(df)
    assert ['A', Path('file.fasta'), 'NA', 'NA'] == df.iloc[0].tolist()
    assert ['B', Path('file2.fasta'), 'NA', 'NA'] == df.iloc[1].tolist()
    assert ['C', 'NA', Path('file_1.fastq'), Path('file_2.fastq')] == df.iloc[2].tolist()

    df = executor.skip_samples_from_input_files(input_samples, skip_samples={'A'})
    df = df.fillna('NA')
    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
    assert 2 == len(df)
    assert ['B', Path('file2.fasta'), 'NA', 'NA'] == df.iloc[0].tolist()
    assert ['C', 'NA', Path('file_1.fastq'), Path('file_2.fastq')] == df.iloc[1].tolist()

    df = executor.skip_samples_from_input_files(input_samples, skip_samples=set('B'))
    df = df.fillna('NA')
    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
    assert 2 == len(df)
    assert ['A', Path('file.fasta'), 'NA', 'NA'] == df.iloc[0].tolist()
    assert ['C', 'NA', Path('file_1.fastq'), Path('file_2.fastq')] == df.iloc[1].tolist()

    df = executor.skip_samples_from_input_files(input_samples, skip_samples={'A', 'C', 'invalid'})
    df = df.fillna('NA')
    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
    assert 1 == len(df)
    assert ['B', Path('file2.fasta'), 'NA', 'NA'] == df.iloc[0].tolist()

    df = executor.skip_samples_from_input_files(input_samples, skip_samples={'A', 'B', 'C'})
    df = df.fillna('NA')
    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
    assert 0 == len(df)


def test_skip_missing_sample_files():
    with tempfile.NamedTemporaryFile() as f1_tmp:
        f1 = Path(f1_tmp.name)

        executor = SnakemakePipelineExecutor()

        input_samples = pd.DataFrame([
            ['A', f1, pd.NA, pd.NA],
            ['B', f1, pd.NA, pd.NA],
            ['C', pd.NA, f1, f1]
        ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

        df = executor.skip_missing_sample_files(input_samples)
        df = df.fillna('NA')
        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
        assert 3 == len(df)
        assert ['A', f1, 'NA', 'NA'] == df.iloc[0].tolist()
        assert ['B', f1, 'NA', 'NA'] == df.iloc[1].tolist()
        assert ['C', 'NA', f1, f1] == df.iloc[2].tolist()

        input_samples = pd.DataFrame([
            ['A', f1, pd.NA, pd.NA],
            ['B', Path('file2.fasta'), pd.NA, pd.NA],
            ['C', pd.NA, Path('file_1.fastq'), Path('file_2.fastq')]
        ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

        df = executor.skip_missing_sample_files(input_samples)
        df = df.fillna('NA')
        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
        assert 1 == len(df)
        assert ['A', f1, 'NA', 'NA'] == df.iloc[0].tolist()

        input_samples = pd.DataFrame([
            ['A', f1, pd.NA, pd.NA],
            ['B', f1, pd.NA, pd.NA],
            ['C', pd.NA, Path('file_1.fastq'), Path('file_2.fastq')]
        ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

        df = executor.skip_missing_sample_files(input_samples)
        df = df.fillna('NA')
        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
        assert 2 == len(df)
        assert ['A', f1, 'NA', 'NA'] == df.iloc[0].tolist()
        assert ['B', f1, 'NA', 'NA'] == df.iloc[1].tolist()

        input_samples = pd.DataFrame([
            ['A', f1, pd.NA, pd.NA],
            ['B', f1, pd.NA, pd.NA],
            ['C', pd.NA, f1, Path('file_2.fastq')]
        ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

        df = executor.skip_missing_sample_files(input_samples)
        df = df.fillna('NA')
        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
        assert 2 == len(df)
        assert ['A', f1, 'NA', 'NA'] == df.iloc[0].tolist()
        assert ['B', f1, 'NA', 'NA'] == df.iloc[1].tolist()

        input_samples = pd.DataFrame([
            ['A', f1, pd.NA, pd.NA],
            ['B', Path('file2.fasta'), pd.NA, pd.NA],
            ['C', pd.NA, f1, f1]
        ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

        df = executor.skip_missing_sample_files(input_samples)
        df = df.fillna('NA')
        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
        assert 2 == len(df)
        assert ['A', f1, 'NA', 'NA'] == df.iloc[0].tolist()
        assert ['C', 'NA', f1, f1] == df.iloc[1].tolist()


def test_validate_input_sample_files():
    input_samples = pd.DataFrame([
        ['A', Path('file.fasta'), pd.NA, pd.NA],
        ['B', Path('file2.fasta'), pd.NA, pd.NA]
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    executor.validate_input_sample_files(input_samples)


def test_validate_input_sample_files_fail():
    input_samples = pd.DataFrame([
        ['A', 'file.fasta', pd.NA, pd.NA],
        ['B', Path('file2.fasta'), pd.NA, pd.NA]
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    with pytest.raises(Exception) as execinfo:
        executor.validate_input_sample_files(input_samples)

    assert 'column=[Assemblies]' in str(execinfo.value)
    assert 'does not contain Path or NA' in str(execinfo.value)


def test_validate_input_sample_files_fail2():
    input_samples = pd.DataFrame([
        ['A', pd.NA, pd.NA, 'file'],
        ['B', Path('file2.fasta'), pd.NA, pd.NA]
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    with pytest.raises(Exception) as execinfo:
        executor.validate_input_sample_files(input_samples)

    assert 'column=[Reads2]' in str(execinfo.value)
    assert 'does not contain Path or NA' in str(execinfo.value)


def test_write_input_sample_files():
    input_samples = pd.DataFrame([
        ['A', Path('file.fasta'), pd.NA, pd.NA],
        ['B', Path('file2.fasta'), pd.NA, pd.NA]
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    with tempfile.TemporaryDirectory() as tmp_dir:
        out_dir = Path(tmp_dir)
        out_file = out_dir / 'output.tsv'

        executor.write_input_sample_files(input_samples, output_file=out_file)

        df = pd.read_csv(out_file, sep='\t').fillna('NA')

        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
        assert 2 == len(df)
        assert ['A', 'file.fasta', 'NA', 'NA'] == df.iloc[0].tolist()
        assert ['B', 'file2.fasta', 'NA', 'NA'] == df.iloc[1].tolist()


def test_write_read_input_sample_files():
    input_samples = pd.DataFrame([
        ['A', Path('file.fasta'), pd.NA, pd.NA],
        ['B', pd.NA, Path('file.fastq'), pd.NA]
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    with tempfile.TemporaryDirectory() as tmp_dir:
        out_dir = Path(tmp_dir)
        out_file = out_dir / 'output.tsv'

        executor.write_input_sample_files(input_samples, output_file=out_file)

        df = executor.read_input_sample_files(out_file)
        executor.validate_input_sample_files(df)

        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == df.columns.tolist()
        assert 2 == len(df)
        assert ['A', Path('file.fasta'), pd.NA, pd.NA] == df.iloc[0].tolist()
        assert ['B', pd.NA, Path('file.fastq'), pd.NA] == df.iloc[1].tolist()


def test_fix_sample_names_no_change():
    input_samples = pd.DataFrame([
        ['A', Path('file1.fasta'), pd.NA, pd.NA],
        ['B', Path('file2.fasta'), pd.NA, pd.NA]
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    input_samples_fixed, original_fixed_names = executor.fix_sample_names(input_samples)
    assert original_fixed_names is None

    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == input_samples_fixed.columns.tolist()
    assert 2 == len(input_samples_fixed)
    assert ['A', 'B'] == input_samples_fixed['Sample'].tolist()
    assert [Path('file1.fasta'), Path('file2.fasta')] == input_samples_fixed['Assemblies'].tolist()


def test_fix_sample_names_with_change():
    input_samples = pd.DataFrame([
        ['A/1', Path('file1.fasta'), pd.NA, pd.NA],
        ['B2', Path('file2.fasta'), pd.NA, pd.NA]
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    input_samples_fixed, original_fixed_names = executor.fix_sample_names(input_samples)
    assert original_fixed_names is not None

    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == input_samples_fixed.columns.tolist()
    assert 2 == len(input_samples_fixed)
    assert ['A__1', 'B2'] == input_samples_fixed['Sample'].tolist()
    assert [Path('file1.fasta'), Path('file2.fasta')] == input_samples_fixed['Assemblies'].tolist()

    assert ['Sample_original', 'Sample_fixed'] == original_fixed_names.columns.tolist()
    assert 2 == len(original_fixed_names)
    assert ['A/1', 'B2'] == original_fixed_names['Sample_original'].tolist()
    assert ['A__1', 'B2'] == original_fixed_names['Sample_fixed'].tolist()


def test_fix_sample_names_with_change2():
    input_samples = pd.DataFrame([
        ['new/A/1/', Path('file1.fasta'), pd.NA, pd.NA],
        ['new/B/2', Path('file2.fasta'), pd.NA, pd.NA],
        ['new/C//2', Path('file3.fasta'), pd.NA, pd.NA]
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    input_samples_fixed, original_fixed_names = executor.fix_sample_names(input_samples)

    assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == input_samples_fixed.columns.tolist()
    assert 3 == len(input_samples_fixed)
    assert ['new__A__1__', 'new__B__2', 'new__C____2'] == input_samples_fixed['Sample'].tolist()
    assert [Path('file1.fasta'), Path('file2.fasta'), Path('file3.fasta')] == input_samples_fixed['Assemblies'].tolist()

    assert ['Sample_original', 'Sample_fixed'] == original_fixed_names.columns.tolist()
    assert 3 == len(original_fixed_names)
    assert ['new/A/1/', 'new/B/2', 'new/C//2'] == original_fixed_names['Sample_original'].tolist()
    assert ['new__A__1__', 'new__B__2', 'new__C____2'] == original_fixed_names['Sample_fixed'].tolist()


def test_fix_sample_names_with_duplicates():
    input_samples = pd.DataFrame([
        ['A/1', Path('file1.fasta'), pd.NA, pd.NA],
        ['A__1', Path('file2.fasta'), pd.NA, pd.NA],
    ], columns=['Sample', 'Assemblies', 'Reads1', 'Reads2'])

    executor = SnakemakePipelineExecutor()

    with pytest.raises(Exception) as execinfo:
        executor.fix_sample_names(input_samples)

    assert "Sanitizing sample names to use as file names leads to duplicates" in str(execinfo.value)


def test_restore_sample_names():
    original_new_names = pd.DataFrame([
        ['new/A/1/', 'new__A__1__'],
        ['new/B/2', 'new__B__2'],
        ['new/C//2', 'new__C____2'],
    ], columns=['Sample_original', 'Sample_fixed'])

    data = pd.DataFrame([
        ['new__A__1__', Path('file1.vcf'), Path('file1.fasta')],
        ['new__B__2', Path('file2.vcf'), Path('file2.fasta')],
        ['new__C____2', Path('file3.vcf'), Path('file3.fasta')]
    ], columns=['Sample', 'VCF', 'Mask'])

    executor = SnakemakePipelineExecutor()

    actual_df = executor.restore_sample_names(data, sample_column='Sample', samples_original_fixed=original_new_names)

    assert ['Sample', 'VCF', 'Mask'] == actual_df.columns.tolist()
    assert 3 == len(actual_df)
    assert ['new/A/1/', 'new/B/2', 'new/C//2'] == actual_df['Sample'].tolist()
    assert [Path('file1.vcf'), Path('file2.vcf'), Path('file3.vcf')] == actual_df['VCF'].tolist()
    assert [Path('file1.fasta'), Path('file2.fasta'), Path('file3.fasta')] == actual_df['Mask'].tolist()



#!/usr/bin/env python
# Converts snippy aligned.fa (and optionally a VCF) to a BED mask file

import sys
from pathlib import Path

from pybedtools import BedTool

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile

if len(sys.argv) == 2:
    snps_aligned = Path(sys.argv[1])
    snps_vcf = None
elif len(sys.argv) == 3:
    snps_aligned = Path(sys.argv[1])
    snps_vcf = Path(sys.argv[2])
else:
    print(f"Usage: {sys.argv[0]} snps.aligned.fa [snps.vcf.gz]\n")
    sys.exit(1)

sf = SequenceFile(snps_aligned)
name, records = list(sf.parse_sequence_file())
regions = MaskedGenomicRegions.from_sequences(records)

if snps_vcf is not None:
    name = name + '.minus-vcf'
    vcf_bed = BedTool(snps_vcf).merge()
    regions = MaskedGenomicRegions(regions.mask.subtract(vcf_bed))

output_file = Path(name + '.bed.gz')
regions.write(output_file)



from typing import List, Any, Optional

import pandas as pd

from genomics_data_index.api.query.features.FeaturesFromIndexComparator import FeatureSamplesSingleCategorySummarizer
from genomics_data_index.api.query.features.FeaturesFromIndexComparator import FeatureSamplesSummarizer
from genomics_data_index.api.query.features.FeaturesFromIndexComparator import FeaturesFromIndexComparator
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet, AllSampleSet
from genomics_data_index.storage.model.db import FeatureSamples, MLSTAllelesSamples
from genomics_data_index.storage.service.MLSTService import MLSTService


class MLSTFeaturesComparator(FeaturesFromIndexComparator):

    def __init__(self, connection: DataIndexConnection,
                 scheme: str = None,
                 locus: str = None,
                 include_unknown_samples: bool = True,
                 include_unknown_no_present_samples=False,
                 include_present: bool = True,
                 include_unknown: bool = False):
        super().__init__(connection=connection, include_unknown_samples=include_unknown_samples,
                         include_unknown_no_present_samples=include_unknown_no_present_samples)
        self._scheme = scheme
        self._locus = locus
        self._include_present = include_present
        self._include_unknown = include_unknown

    def _get_total(self, samples: SampleSet) -> int:
        if self._scheme is None:
            return super()._get_total(samples)
        else:
            if isinstance(samples, AllSampleSet):
                return self._connection.sample_service.count_samples_associated_with_mlst_scheme(self._scheme)
            else:
                return len(samples)

    @property
    def summary_columns(self) -> List[str]:
        return self.feature_id_columns + FeatureSamplesSingleCategorySummarizer.SUMMARY_NAMES

    @property
    def feature_id_columns(self) -> List[str]:
        return ['Scheme', 'Locus', 'Allele']

    @property
    def index_name(self) -> str:
        return 'MLST Feature'

    def _create_feature_sample_count_row(self, feature_id: str, feature: FeatureSamples,
                                         feature_samples: SampleSet,
                                         feature_samples_unknown: Optional[SampleSet],
                                         total: int,
                                         feature_samples_summarizer: FeatureSamplesSummarizer) -> List[Any]:
        if isinstance(feature, MLSTAllelesSamples):
            summary_data = feature_samples_summarizer.summary_data(samples=feature_samples, total=total,
                                                                   unknown_samples=feature_samples_unknown)
            return [feature.query_id, feature.scheme, feature.locus, feature.allele] + summary_data
        else:
            raise Exception(f'feature={feature} is not of type {MLSTAllelesSamples.__name__}')

    def _do_summary(self, sample_set: SampleSet, feature_samples_summarizer: FeatureSamplesSummarizer) -> pd.DataFrame:
        mlst_service: MLSTService = self._connection.mlst_service
        mlst_present_features = mlst_service.get_features(scheme=self._scheme,
                                                          locus=self._locus,
                                                          include_present=self._include_present,
                                                          include_unknown=self._include_unknown)
        return self._create_summary_df(mlst_present_features, present_samples=sample_set,
                                       feature_samples_summarizer=feature_samples_summarizer)



import logging
from typing import Generator, Set

from genomics_data_index.storage.io.FeaturesReader import FeaturesReader
from genomics_data_index.storage.io.SampleData import SampleData
from genomics_data_index.storage.io.SampleDataPackage import SampleDataPackage
from genomics_data_index.storage.io.mlst.MLSTFeaturesReader import MLSTFeaturesReader
from genomics_data_index.storage.io.mlst.MLSTSampleData import MLSTSampleData

logger = logging.getLogger(__name__)


class MLSTSampleDataPackage(SampleDataPackage):

    def __init__(self, mlst_reader: MLSTFeaturesReader, index_unknown_missing: bool = True):
        super().__init__(index_unknown_missing=index_unknown_missing)
        self._mlst_reader = mlst_reader

    def get_features_reader(self) -> FeaturesReader:
        if not self.index_unknown_missing():
            logger.warning(f'index_unknown_missing=False for MLSTSampleDataPackage, but right now '
                           f'missing data for MLST results is always indexed so this will be ignored.')
        return self._mlst_reader

    def sample_names(self) -> Set[str]:
        return set(self._mlst_reader.samples_list())

    def process_all_data(self) -> SampleDataPackage:
        processed_data = {}
        for sample_data in self.iter_sample_data():
            processed_data[sample_data.sample_name] = sample_data
        return self

    def iter_sample_data(self) -> Generator[SampleData, None, None]:
        for sample_name, scheme, data in self._mlst_reader.iter_sample_data():
            yield MLSTSampleData(sample_name=sample_name, scheme=scheme, mlst_allele_data=data)



import abc
from typing import Union, Iterable, Tuple, Set, Dict

from ete3 import Tree, TreeStyle, Face, RectFace, CircleFace, TextFace

from genomics_data_index.api.query.SamplesQuery import SamplesQuery


class TreeSamplesVisual(abc.ABC):
    """
    Defines a visual element on a tree which applies to a set of samples (e.g., highlight or annotate).
    """

    LEGEND_FACE_KINDS = ['circle', 'circ', 'c',
                         'rectangle', 'rect', 'r']

    def __init__(self, samples: Union[SamplesQuery, Iterable[str]],
                 legend_nodesize: int,
                 legend_fontsize: int,
                 legend_columns: Dict[str, int]):
        """
        Creates a new TreeSamplesVisual with the given information.
        :param samples: The set of samples to apply this visual style to.
        :param legend_nodesize: The legend node size for this set of samples.
        :param legend_fontsize: The legend font size for this set of samples.
        :param legend_columns: A dict mapping legend item types to column numbers.
        :return: A new TreeSamplesVisual.
        """
        self._samples = samples
        self._legend_nodesize = legend_nodesize
        self._legend_fontsize = legend_fontsize
        self._present_sample_names = None
        self._unknown_sample_names = None
        self._legend_columns = legend_columns

    @property
    def present_sample_names(self) -> Set[str]:
        """
        Gets the names of the samples present in the selected set this visual style applies to.
        :return: A set of the names of present samples this visual style applies to.
        """
        if self._present_sample_names is None:
            if isinstance(self._samples, SamplesQuery):
                self._present_sample_names = set(self._samples.tolist(names=True))
            else:
                self._present_sample_names = set(self._samples)
        return self._present_sample_names

    @property
    def unknown_sample_names(self) -> Set[str]:
        """
        Gets the names of the samples unknown in the selected set this visual style applies to.
        :return: A set of the names of unknown samples this visual style applies to.
        """
        if self._unknown_sample_names is None:
            if isinstance(self._samples, SamplesQuery):
                self._unknown_sample_names = set(self._samples.select_unknown().tolist(names=True))
            else:
                # If we don't have a SamplesQuery, there are no unknown samples
                self._unknown_sample_names = set()
        return self._unknown_sample_names

    @abc.abstractmethod
    def apply_visual(self, tree: Tree, tree_style: TreeStyle) -> None:
        """
        Applies a visual style defined by the passed samples to the Tree and TreeStyle.
        Will modify tree and tree_style in-place (this avoids having to copy large trees
        which was causing me issues).

        :param tree: The Tree to apply visual information to.
        :param tree_style: The TreeStyle to apply visual information to.
        :returns: None. Modifies tree and tree_style in-place.
        """
        pass

    def _add_legend_entry(self, legend_label: str, legend_color: str, unknown_color: str,
                          include_unknown: bool, has_unknowns: bool, kind: str, tree_style: TreeStyle) -> None:
        if legend_label is not None:
            color_face, unknown_face, text_face = self._build_legend_item(color=legend_color,
                                                                          unknown_color=unknown_color,
                                                                          include_unknown=include_unknown,
                                                                          has_unknowns=has_unknowns,
                                                                          legend_label=legend_label,
                                                                          kind=kind)

            tree_style.legend.add_face(color_face, column=self._legend_columns['present'])
            if include_unknown:
                tree_style.legend.add_face(unknown_face, column=self._legend_columns['unknown'])
            tree_style.legend.add_face(text_face, column=self._legend_columns['text'])

    def _build_legend_item(self, color: str, unknown_color: str, include_unknown: bool, has_unknowns: bool,
                           legend_label: str, kind: str) -> Tuple[Face, Face, Face]:
        if kind == 'rect' or kind == 'rectangle' or kind == 'r':
            cf = RectFace(width=self._legend_nodesize, height=self._legend_nodesize, bgcolor=color,
                          fgcolor='black')
            if include_unknown:
                if has_unknowns:
                    ucf = RectFace(width=self._legend_nodesize, height=self._legend_nodesize, bgcolor=unknown_color,
                                   fgcolor='black')
                else:
                    ucf = RectFace(width=self._legend_nodesize, height=self._legend_nodesize, bgcolor=None,
                                   fgcolor=None)
            else:
                ucf = None
        elif kind == 'circle' or kind == 'circ' or kind == 'c':
            cf = CircleFace(radius=self._legend_nodesize / 2, color=color)
            if include_unknown:
                if has_unknowns:
                    ucf = CircleFace(radius=self._legend_nodesize / 2, color=unknown_color)
                else:
                    ucf = CircleFace(radius=self._legend_nodesize / 2, color=None)
            else:
                ucf = None
        else:
            raise Exception(f'kind=[{kind}] must be one of {self.LEGEND_FACE_KINDS}')
        cf.hz_align = 2
        cf.margin_left = 3
        cf.margin_right = 3
        cf.margin_top = 3
        cf.margin_bottom = 3
        if ucf is not None:
            ucf.hz_align = 2
            ucf.margin_left = 3
            ucf.margin_right = 3
            ucf.margin_top = 3
            ucf.margin_bottom = 3
        tf = TextFace(legend_label, fsize=self._legend_fontsize)
        tf.margin_left = 10
        tf.margin_right = 10
        return cf, ucf, tf



import abc
from typing import Set

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.FeaturesReader import FeaturesReader


class NucleotideFeaturesReader(FeaturesReader):

    def __init__(self):
        super().__init__()

    def _minimal_expected_columns(self) -> Set[str]:
        return {'SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE'}

    @abc.abstractmethod
    def get_genomic_masked_region(self, sample_name: str) -> MaskedGenomicRegions:
        pass



import pytest

from genomics_data_index.storage.util.ListSliceIter import ListSliceIter


def test_slice_list():
    # Case 3 elements
    data = [1, 2, 3]
    assert [[1], [2], [3]] == list(ListSliceIter(data, slice_size=1).islice())
    assert [[1, 2], [3]] == list(ListSliceIter(data, slice_size=2).islice())
    assert [[1, 2, 3]] == list(ListSliceIter(data, slice_size=3).islice())
    assert [[1, 2, 3]] == list(ListSliceIter(data, slice_size=4).islice())

    # Trying slice < 1
    with pytest.raises(Exception) as execinfo:
        ListSliceIter(data, slice_size=0)
    assert 'must be a positive integer' in str(execinfo.value)

    # Case empty data
    data = []
    assert [] == list(ListSliceIter(data, slice_size=1).islice())
    assert [] == list(ListSliceIter(data, slice_size=2).islice())

    # Case 5 elements and strings
    data = ['A', 'B', 'C', 'D', 'E']
    assert [['A'], ['B'], ['C'], ['D'], ['E']] == list(ListSliceIter(data, slice_size=1).islice())
    assert [['A', 'B'], ['C', 'D'], ['E']] == list(ListSliceIter(data, slice_size=2).islice())
    assert [['A', 'B', 'C'], ['D', 'E']] == list(ListSliceIter(data, slice_size=3).islice())
    assert [['A', 'B', 'C', 'D'], ['E']] == list(ListSliceIter(data, slice_size=4).islice())
    assert [['A', 'B', 'C', 'D', 'E']] == list(ListSliceIter(data, slice_size=5).islice())
    assert [['A', 'B', 'C', 'D', 'E']] == list(ListSliceIter(data, slice_size=6).islice())



import math
from typing import List, Any, Generator


class ListSliceIter:

    def __init__(self, data: List[Any], slice_size: int = 1):
        if not isinstance(data, list):
            raise Exception(f'data={data} is not of type list')
        if not isinstance(slice_size, int) or slice_size <= 0:
            raise Exception(f'slice_size={slice_size} must be a positive integer')

        self._data = data
        self._slice_size = slice_size

    def islice(self) -> Generator[List[Any], None, None]:
        num_chunks = math.ceil(len(self._data) / self._slice_size)
        for chunk in range(num_chunks):
            start = chunk * self._slice_size
            end = min(start + self._slice_size, len(self._data))
            yield self._data[start:end]



import logging
import os
import subprocess
from pathlib import Path

import pandas as pd

logger = logging.getLogger(__name__)


class FilesystemStorage:
    ALL_SUBDIRECTORIES = ['reference', 'kmer', 'variation', 'mlst']

    def __init__(self, root_dir: Path):
        self._root_dir = root_dir

        if not root_dir.exists():
            os.mkdir(root_dir)

    def _check_make_dir(self, name) -> Path:
        d = self._root_dir / name
        if not d.exists():
            os.mkdir(d)
        return d

    def get_storage_size(self) -> pd.DataFrame:
        '''
        Gets the total size of all files on the disk in bytes.
        :return: A Dataframe of the size of all files on the disk (in bytes).
        '''

        # There is likely a Python-specific way of doing this but I haven't found a simple solution
        # This thread (https://stackoverflow.com/questions/1392413/calculating-a-directorys-size-using-python/1392549)
        # has a lot of different implementations but they all have complexities with symlinks, subdirectories,
        # broken links, etc.
        logger.warning(
            'A reminder to myself to look for a Python solution for directory sizes (instead of running `du`)')

        sizes_list = []
        for dname in self.ALL_SUBDIRECTORIES:
            dpath = self._check_make_dir(dname)
            # Line from https://stackoverflow.com/a/47930319
            number_of_files = file_count = sum(len(files) for _, _, files in os.walk(dpath))
            size = subprocess.check_output(['du', '-s', '--block-size=1', dpath]).split()[0].decode('utf-8')
            sizes_list.append(['Filesystem',
                               self._root_dir.name,
                               dpath.name,
                               int(size),
                               number_of_files])

        return pd.DataFrame(sizes_list, columns=['Type', 'Name', 'Division', 'Data Size', 'Number of Items'])

    @property
    def root_dir(self):
        return self._root_dir

    @property
    def reference_dir(self):
        return self._check_make_dir('reference')

    @property
    def kmer_dir(self):
        return self._check_make_dir('kmer')

    @property
    def variation_dir(self):
        return self._check_make_dir('variation')

    @property
    def mlst_dir(self):
        return self._check_make_dir('mlst')



from __future__ import annotations

import abc
from typing import List, Optional


class QueryFeature(abc.ABC):
    WILD = '*'
    SPLIT_CHAR = ':'

    def __init__(self):
        pass

    @property
    def id(self) -> Optional[str]:
        return f'{self.prefix}{self.id_no_prefix}'

    @property
    @abc.abstractmethod
    def id_no_prefix(self) -> Optional[str]:
        pass

    def is_wild(self) -> bool:
        return self.WILD in self.id

    @abc.abstractmethod
    def is_unknown(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def scope(self) -> str:
        pass

    @abc.abstractmethod
    def to_unknown(self) -> QueryFeature:
        """
        Converts this given QueryFeature to one representing an unknown (e.g., unknown mutation or MLST allele).
        :return: The equivalent of this feature but representing an unknown.
        """
        pass

    @abc.abstractmethod
    def to_unknown_explode(self) -> List[QueryFeature]:
        """
        Converts this given QueryFeature to one representing an unknown (e.g., unknown mutation or MLST allele).
        This will explode into separate unknown features (e.g., in cases of an indel/complex mutation).
        :return: A list of unknown features, where each feature corresponds to a single position.
        """
        pass

    @property
    def prefix(self) -> str:
        return ''

    def __str__(self) -> str:
        return self.id

    def __repr__(self) -> str:
        return self.__str__()



import logging
import tempfile
from pathlib import Path

import pytest
from pybedtools import BedTool

from genomics_data_index.storage.MaskedGenomicRegions import MaskedGenomicRegions
from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData
from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile
from genomics_data_index.test.integration import snippy_sample_vcfs_dict, snippy_sample_mask_sequences_dict
from genomics_data_index.test.integration import snpeff_sample_vcfs

logger = logging.getLogger(__name__)


@pytest.fixture
def sample_dataA():
    tmp_file = Path(tempfile.mktemp())
    seq_name, seq_file = SequenceFile(snippy_sample_mask_sequences_dict['SampleA']).parse_sequence_file()
    mask_region = MaskedGenomicRegions.from_sequences(seq_file)
    mask_region.write(tmp_file)
    sample_data = NucleotideSampleData(sample_name='SampleA',
                                       vcf_file=snippy_sample_vcfs_dict['SampleA'],
                                       vcf_file_index=None,
                                       mask_bed_file=tmp_file,
                                       preprocessed=True)
    return sample_data


@pytest.fixture
def sample_data_sh10_014():
    tmp_file = Path(tempfile.mktemp())
    mask_region = MaskedGenomicRegions(BedTool([('NC_011083', 0, 100), ('NC_011083', 199, 250)]))
    mask_region.write(tmp_file)
    sample_data = NucleotideSampleData(sample_name='SH10-014',
                                       vcf_file=snpeff_sample_vcfs['SH10-014'],
                                       vcf_file_index=None,
                                       mask_bed_file=tmp_file,
                                       preprocessed=True)
    return sample_data


def test_read_sample_data_features_no_unknown(sample_dataA):
    df = sample_dataA.read_sample_data_features(include_masked_regions=False).sort_values('POS')

    assert 46 == len(df)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(df.columns)
    assert {'SampleA'} == set(df['SAMPLE'].tolist())

    # Variant position
    v = df[df['POS'] == 461]
    assert ['reference'] == v['CHROM'].tolist()
    assert ['AAAT'] == v['REF'].tolist()
    assert ['G'] == v['ALT'].tolist()
    assert ['reference:461:AAAT:G'] == v['VARIANT_ID'].tolist()

    # Variant position
    v = df[df['POS'] == 1048]
    assert ['reference'] == v['CHROM'].tolist()
    assert ['C'] == v['REF'].tolist()
    assert ['G'] == v['ALT'].tolist()
    assert ['reference:1048:C:G'] == v['VARIANT_ID'].tolist()

    # No snpeff annotations
    assert all(df['ANN.Annotation'].isna())

    # No missing/unknown
    assert not any(df['ALT'] == '?')


def test_read_sample_data_features_with_unknown(sample_dataA):
    df = sample_dataA.read_sample_data_features(include_masked_regions=True).sort_values('POS')

    assert 482 == len(df)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(df.columns)
    assert {'SampleA'} == set(df['SAMPLE'].tolist())

    # Number actual variants
    assert 45 == len(df[df['ALT'] != '?'])

    # Number missing
    assert 437 == len(df[df['ALT'] == '?'])

    # 461... is missing
    assert ['reference:460:1:?', 'reference:461:1:?',
            'reference:462:1:?'] == df[
               (df['POS'] >= 460) & (df['POS'] <= 465)]['VARIANT_ID'].tolist()
    logger.warning('I need to verify that mask coordinates are correct')

    # Variant position
    v = df[df['POS'] == 1048]
    assert ['reference'] == v['CHROM'].tolist()
    assert ['C'] == v['REF'].tolist()
    assert ['G'] == v['ALT'].tolist()
    assert ['reference:1048:C:G'] == v['VARIANT_ID'].tolist()

    # No snpeff annotations
    assert all(df['ANN.Annotation'].isna())

    # One of the missing
    assert ['reference:1:1:?'] == df[df['POS'] == 1]['VARIANT_ID'].tolist()


def test_read_sample_data_features_snpeff_with_unknown(sample_data_sh10_014):
    df = sample_data_sh10_014.read_sample_data_features(include_masked_regions=True).sort_values('POS')

    num_variants = 139
    num_masked = 151
    assert num_variants + num_masked == len(df)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(df.columns)
    assert {'SH10-014'} == set(df['SAMPLE'].tolist())

    # Number actual variants
    assert num_variants == len(df[df['ALT'] != '?'])

    # Number missing
    assert num_masked == len(df[df['ALT'] == '?'])

    variant_ids = set(df['VARIANT_ID'].tolist())

    # Test to verify some are mossing
    assert 'NC_011083:1:1:?' in variant_ids
    assert 'NC_011083:100:1:?' in variant_ids
    assert 'NC_011083:101:1:?' not in variant_ids
    assert 'NC_011083:199:1:?' not in variant_ids
    assert 'NC_011083:200:1:?' in variant_ids
    assert 'NC_011083:250:1:?' in variant_ids
    assert 'NC_011083:251:1:?' not in variant_ids

    # Variant position
    v = df[df['POS'] == 140658]
    assert ['NC_011083'] == v['CHROM'].tolist()
    assert ['C'] == v['REF'].tolist()
    assert ['A'] == v['ALT'].tolist()
    assert ['NC_011083:140658:C:A'] == v['VARIANT_ID'].tolist()
    assert ['missense_variant'] == v['ANN.Annotation'].tolist()
    assert ['murF'] == v['ANN.Gene_Name'].tolist()
    assert ['c.497C>A'] == v['ANN.HGVS.c'].tolist()
    assert ['p.Ala166Glu'] == v['ANN.HGVS.p'].tolist()


def test_combine_vcf_mask(sample_dataA: NucleotideSampleData):
    expected_columns = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
                        'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                        'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

    sampleA_vcf_df = sample_dataA.read_vcf_features()
    assert 46 == len(sampleA_vcf_df)
    assert 461 in sampleA_vcf_df['POS'].tolist()

    sampleA_mask_df = sample_dataA.get_mask().mask_to_features()
    assert 437 == len(sampleA_mask_df)
    assert 461 in sampleA_mask_df['POS'].tolist()

    combined_df = sample_dataA.combine_vcf_mask(vcf_frame=sampleA_vcf_df, mask_frame=sampleA_mask_df)

    assert expected_columns == combined_df.columns.tolist()
    assert 482 == len(combined_df)
    assert 437 == len(combined_df[combined_df['TYPE'] == 'UNKNOWN_MISSING'])
    assert 45 == len(combined_df[combined_df['TYPE'] != 'UNKNOWN_MISSING'])
    logger.warning('I need to verify that mask coordinates are correct')

    # Make sure it's the SNV/SNP from the VCF file that I keep and not the missing position in this file
    # assert combined_df[combined_df['POS'] == 461]['TYPE'] == 'INDEL'
    assert ['UNKNOWN_MISSING'] == combined_df[combined_df['POS'] == 461]['TYPE'].tolist()



from __future__ import annotations

from typing import Union, List

import pandas as pd
from ete3 import Tree

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl.TreeSamplesQueryFactory import TreeSamplesQueryFactory
from genomics_data_index.api.query.impl.WrappedSamplesQuery import WrappedSamplesQuery
from genomics_data_index.configuration.connector import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet


class DataFrameSamplesQuery(WrappedSamplesQuery):
    """
    Defines a SamplesQuery which is joined to a DataFrame. This allows for queries that make use of data in the DataFrame.
    """

    ISIN_KINDS = ['dataframe']
    ISA_KINDS = ['dataframe']

    def __init__(self, connection: DataIndexConnection, wrapped_query: SamplesQuery,
                 data_frame: pd.DataFrame,
                 sample_ids_col: str,
                 default_isa_kind: str,
                 default_isa_column: str):
        """
        Builds a new DataFrameSamplesQuery from the given information. In most normal operations DataFrameSamplesQuery objects
        are not created directly but are instead created from an :py:class:`genomics_data_index.api.GenomicsDataIndex`
        object or from the join operation applied to a SamplesQuery.

        :param connection: A connection to a database containing samples.
        :param wrapped_query: The SamplesQuery to wrap around/decorate.
        :param data_frame: The DataFrame to join to this query. This dataframe must have a column which matches rows to
                           the Sample IDs stored in database defined by the connection object.
        :param sample_ids_col: The name of the column in the DataFrame defining the Sample ID for a particular row.
        :param default_isa_kind: If A is a DataFrameSamplesQuery defines the default `kind` parameter to be used
                                 when running A.isa(data, kind='...').
        :param default_isa_column: If A is a DataFrameSamplesQuery defines the default 'isa_column' parameter
                                   to be used when running A.isa(data, kind='...', isa_column='...').
        :return: A new DataFrameSamplesQuery object.
        """
        super().__init__(connection=connection, wrapped_query=wrapped_query)
        self._sample_ids_col = sample_ids_col
        self._data_frame = data_frame
        self._default_isa_kind = default_isa_kind
        self._default_isa_column = default_isa_column

    def _isin_internal(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet], kind: str,
                       **kwargs) -> SamplesQuery:
        if kind == 'dataframe':
            if isinstance(data, pd.Series) and data.dtype == bool:
                if data.index.equals(self._data_frame.index):
                    return self._handle_select_by_series(data, query_message='isin(subset from series)')
                else:
                    raise Exception(f'Passed data=[series with length {len(data)}] '
                                    f'does not have same index as internal data frame (length={len(self._data_frame)})')
            else:
                raise Exception(f'data=[{data}] is wrong type for kind=[{kind}]. '
                                f'Must be a boolean pandas.Series')
        else:
            raise Exception(f'kind=[{kind}] is not supported. Must be one of {self._isin_kinds()}')

    def _isin_kinds(self) -> List[str]:
        return list(set(super()._isin_kinds() + self.ISIN_KINDS))

    def _can_handle_isin_kind(self, kind: str) -> bool:
        return kind in self.ISIN_KINDS

    def _can_handle_isa_kind(self, kind: str) -> bool:
        return kind in self.ISA_KINDS

    def isa(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet], kind: str = None,
            **kwargs) -> SamplesQuery:
        if kind is None:
            if self._default_isa_kind is None:
                kind = 'sample'
            else:
                kind = self._default_isa_kind

        return super().isa(data=data, kind=kind, **kwargs)

    def features_comparison(self, sample_categories: Union[List[SamplesQuery], List[SampleSet], str],
                            category_prefixes: List[str] = None,
                            categories_kind: str = 'samples',
                            kind: str = 'mutations',
                            unit: str = 'percent',
                            category_samples_threshold: int = None,
                            include_unknown_samples: bool = False, include_unknown_no_present_samples: bool = False,
                            use_only_samples_in_categories: bool = True,
                            **kwargs) -> pd.DataFrame:
        if categories_kind == 'dataframe':
            if not isinstance(sample_categories, str):
                raise Exception(f'sample_categories={sample_categories} is type {type(sample_categories)}. '
                                f'For categories_kind={categories_kind}, sample_categories={sample_categories} '
                                f'must be a string')
            elif sample_categories not in self._data_frame:
                raise Exception(f'sample_categories={sample_categories} is not one of the columns of the dataframe. '
                                f'For categories_kind={categories_kind}, sample_categories refers to a column in the '
                                f'dataframe, which must be one of {self._data_frame.columns.tolist()}')

            aggregated_df = self.toframe().groupby(sample_categories).agg({
                'Sample ID': SampleSet,
            })
            aggregated_df = aggregated_df.sort_index()
            categories = aggregated_df['Sample ID'].tolist()  # list of SampleSets
            category_prefixes = aggregated_df.index.tolist()  # List of category prefixes

            return self._wrapped_query.features_comparison(sample_categories=categories,
                                                           category_prefixes=category_prefixes,
                                                           categories_kind='samples',
                                                           kind=kind,
                                                           unit=unit,
                                                           category_samples_threshold=category_samples_threshold,
                                                           include_unknown_samples=include_unknown_samples,
                                                           include_unknown_no_present_samples=include_unknown_no_present_samples,
                                                           use_only_samples_in_categories=use_only_samples_in_categories,
                                                           **kwargs)
        else:
            return super().features_comparison(sample_categories=sample_categories,
                                               category_prefixes=category_prefixes,
                                               categories_kind=categories_kind,
                                               kind=kind,
                                               unit=unit,
                                               category_samples_threshold=category_samples_threshold,
                                               include_unknown_samples=include_unknown_samples,
                                               include_unknown_no_present_samples=include_unknown_no_present_samples,
                                               use_only_samples_in_categories=use_only_samples_in_categories,
                                               **kwargs)

    def _isa_internal(self, data: Union[str, List[str], pd.Series, SamplesQuery, SampleSet], kind: str,
                      isa_column: str = None,
                      regex: bool = False) -> SamplesQuery:
        if kind == 'dataframe':
            if isa_column is None and self._default_isa_column is not None:
                isa_column = self._default_isa_column
            elif isa_column is None:
                raise Exception(f'No defined isa_column, cannot execute isa for kind={kind}')

            if regex:
                no_na_select = (~self._data_frame[isa_column].isna()) & (
                    self._data_frame[isa_column].str.contains(data))
                return self._handle_select_by_series(no_na_select,
                                                     query_message=f"isa('{isa_column}' contains '{data}')")
            else:
                return self._handle_select_by_series(self._data_frame[isa_column] == data,
                                                     query_message=f"isa('{isa_column}' is '{data}')")
        else:
            raise Exception(f'Invalid kind={kind}. Must be one of {self._isa_kinds()}')

    def _isa_kinds(self):
        return super()._isa_kinds() + self.ISA_KINDS

    def _handle_select_by_series(self, series_selection: pd.Series, query_message: str) -> SamplesQuery:
        subset_df = self._data_frame[series_selection]
        subset_sample_set = SampleSet(subset_df[self._sample_ids_col].tolist())
        subset_query = self._wrapped_query.intersect(sample_set=subset_sample_set, query_message=query_message)
        return self._wrap_create(subset_query)

    def toframe(self, include_present: bool = True, include_unknown: bool = False,
                include_absent: bool = False) -> pd.DataFrame:
        samples_dataframe = super().toframe(include_present=include_present,
                                            include_unknown=include_unknown,
                                            include_absent=include_absent)
        samples_df_cols = list(samples_dataframe.columns)
        merged_df = self._data_frame.merge(samples_dataframe, how='inner', left_on=self._sample_ids_col,
                                           right_on='Sample ID')
        new_col_order = samples_df_cols + [col for col in list(merged_df.columns) if col not in samples_df_cols]
        return merged_df[new_col_order]

    def reset_universe(self, include_unknown: bool = True) -> SamplesQuery:
        new_wrapped_query = self._wrapped_query.reset_universe(include_unknown=include_unknown)
        return self._wrap_create(new_wrapped_query)

    def _wrap_create(self, wrapped_query: SamplesQuery) -> WrappedSamplesQuery:
        return DataFrameSamplesQuery(connection=self._query_connection,
                                     wrapped_query=wrapped_query,
                                     data_frame=self._data_frame,
                                     sample_ids_col=self._sample_ids_col,
                                     default_isa_kind=self._default_isa_kind,
                                     default_isa_column=self._default_isa_column)

    def build_tree(self, kind: str, **kwargs) -> SamplesQuery:
        return TreeSamplesQueryFactory.instance().build_tree(kind=kind,
                                                             database_connection=self._query_connection,
                                                             wrapped_query=self, **kwargs)

    def join_tree(self, tree: Tree, kind='mutation', **kwargs) -> SamplesQuery:
        return TreeSamplesQueryFactory.instance().join_tree(tree=tree,
                                                            kind=kind,
                                                            database_connection=self._query_connection,
                                                            wrapped_query=self,
                                                            **kwargs)

    def join(self, data_frame: pd.DataFrame, sample_ids_column: str = None,
             sample_names_column: str = None, default_isa_kind: str = 'names',
             default_isa_column: str = None) -> SamplesQuery:
        raise Exception(f'Cannot join a new dataframe onto an existing data frame query: {self}')

    @classmethod
    def create_with_sample_ids_column(self, sample_ids_column: str, data_frame: pd.DataFrame,
                                      wrapped_query: SamplesQuery, connection: DataIndexConnection,
                                      query_message: str = None,
                                      default_isa_kind: str = None,
                                      default_isa_column: str = None) -> DataFrameSamplesQuery:
        """
        Builds a new DataFrameSamplesQuery from the given information matching by a column defining the Sample IDs.
        In most normal operations DataFrameSamplesQuery objects are not created directly but are instead created from
        an :py:class:`genomics_data_index.api.GenomicsDataIndex` object or from the join operation applied to a SamplesQuery.

        :param sample_ids_col: The name of the column in the DataFrame defining the Sample ID for a particular row.
        :param data_frame: The DataFrame to join to this query. This dataframe must have a column which matches rows to
                   the Sample IDs stored in database defined by the connection object.
        :param wrapped_query: The SamplesQuery to wrap around/decorate.
        :param connection: A connection to a database containing samples.
        :param query_message: The query_message to use when creating this new DataFrameSamplesQuery object.
        :param default_isa_kind: If A is a DataFrameSamplesQuery defines the default `kind` parameter to be used
                                 when running A.isa(data, kind='...').
        :param default_isa_column: If A is a DataFrameSamplesQuery defines the default 'isa_column' parameter
                                   to be used when running A.isa(data, kind='...', isa_column='...').
        :return: A new DataFrameSamplesQuery object.
        """
        sample_ids = data_frame[sample_ids_column].tolist()
        df_sample_set = SampleSet(sample_ids=sample_ids)
        universe_set = wrapped_query.universe_set.intersection(df_sample_set)

        if query_message is None:
            query_message = f'dataframe(ids_col=[{sample_ids_column}])'

        wrapped_query_intersect = wrapped_query.set_universe(universe_set, query_message=query_message)

        return DataFrameSamplesQuery(connection=connection,
                                     wrapped_query=wrapped_query_intersect,
                                     data_frame=data_frame,
                                     sample_ids_col=sample_ids_column,
                                     default_isa_column=default_isa_column,
                                     default_isa_kind=default_isa_kind)

    @classmethod
    def create_with_sample_names_column(self, sample_names_column: str, data_frame: pd.DataFrame,
                                        wrapped_query: SamplesQuery,
                                        connection: DataIndexConnection,
                                        default_isa_kind: str = None,
                                        default_isa_column: str = None) -> DataFrameSamplesQuery:
        """
        Builds a new DataFrameSamplesQuery from the given information matching by a column defining the Sample names.
        In most normal operations DataFrameSamplesQuery objects are not created directly but are instead created from
        an :py:class:`genomics_data_index.api.GenomicsDataIndex` object or from the join operation applied to a SamplesQuery.

        :param sample_names_column: The name of the column in the DataFrame defining the Sample name for a particular row.
                                    This will internally be matched up to Sample IDs in the database defined by the connection
                                    object and used to create a new column in the DataFrame with the Sample ID.
        :param data_frame: The DataFrame to join to this query. This dataframe must have a column which matches rows to
                   the Sample names stored in database defined by the connection object.
        :param wrapped_query: The SamplesQuery to wrap around/decorate.
        :param connection: A connection to a database containing samples.
        :param default_isa_kind: If A is a DataFrameSamplesQuery defines the default `kind` parameter to be used
                                 when running A.isa(data, kind='...').
        :param default_isa_column: If A is a DataFrameSamplesQuery defines the default 'isa_column' parameter
                                   to be used when running A.isa(data, kind='...', isa_column='...').
        :return: A new DataFrameSamplesQuery object.
        """
        sample_names = set(data_frame[sample_names_column].tolist())
        sample_ids_column = 'Sample ID'

        sample_name_ids = connection.sample_service.find_sample_name_ids(sample_names)

        # Only attempt once to rename sample IDs column if it already exists
        if sample_ids_column in data_frame:
            sample_ids_column = sample_ids_column + '_gdi'
            if sample_ids_column in data_frame:
                raise Exception(f'Column to be used for sample_ids [{sample_ids_column}] already in data frame')

        sample_ids_series = pd.Series(sample_name_ids, name=sample_ids_column)
        data_frame = data_frame.merge(sample_ids_series, left_on=sample_names_column, right_index=True)

        query_message = f'dataframe(names_col=[{sample_names_column}])'

        return DataFrameSamplesQuery.create_with_sample_ids_column(sample_ids_column=sample_ids_column,
                                                                   data_frame=data_frame,
                                                                   wrapped_query=wrapped_query,
                                                                   connection=connection,
                                                                   query_message=query_message,
                                                                   default_isa_column=default_isa_column,
                                                                   default_isa_kind=default_isa_kind)



from pathlib import Path
from typing import List, Dict, Iterable

import ga4gh.vrs.dataproxy as dataproxy
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from biocommons.seqrepo import SeqRepo
from ete3 import Tree
from sqlalchemy.exc import NoResultFound

from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI
from genomics_data_index.storage.model.db import Reference, SampleNucleotideVariation, ReferenceSequence, Sample
from genomics_data_index.storage.service import DatabaseConnection, EntityExistsError


class ReferenceService:
    MUTATION_ID_TYPES = ['spdi_ref']

    def __init__(self, database_connection: DatabaseConnection, seq_repo_dir: Path,
                 seq_repo_namespace: str = 'genomics_data_index'):
        self._connection = database_connection
        self._seq_repo_namespace = seq_repo_namespace

        if seq_repo_dir is not None:
            self._seq_repo_updatable = SeqRepo(seq_repo_dir, writeable=True)
            self._seq_repo_proxy = dataproxy.SeqRepoDataProxy(SeqRepo(seq_repo_dir))

    def add_reference_genome(self, genome_file: Path):
        (genome_name, sequences) = SequenceFile(genome_file).parse_sequence_file()

        if self.exists_reference_genome(genome_name):
            raise EntityExistsError(f'Reference genome [{genome_name}] already exists in database')
        else:
            for record in sequences:
                self._seq_repo_updatable.store(str(record.seq),
                                               [{'namespace': self._seq_repo_namespace, 'alias': record.id}])
            self._seq_repo_updatable.commit()

            self._create_reference_genome_db(genome_file)

    def get_reference_sequences(self, reference_name: str) -> Dict[str, ReferenceSequence]:
        reference = self.find_reference_genome(reference_name)
        return {s.sequence_name: s for s in reference.sequences}

    def get_sequence(self, sequence_name: str) -> SeqRecord:
        namespace = self._seq_repo_namespace
        seq_string = self._seq_repo_proxy.get_sequence(f'{namespace}:{sequence_name}')
        return SeqRecord(Seq(seq_string), id=sequence_name)

    def get_reference_genome_records(self, reference_name: str) -> List[SeqRecord]:
        reference = self.find_reference_genome(reference_name)
        return [self.get_sequence(sequence.sequence_name) for sequence in reference.sequences]

    def _create_reference_genome_db(self, reference_file: Path):
        ref_length = 0
        ref_contigs = {}

        (ref_name, sequences) = SequenceFile(reference_file).parse_sequence_file()
        for record in sequences:
            ref_contigs[record.id] = ReferenceSequence(
                sequence_name=record.id, sequence_length=len(record.seq))
            ref_length += len(record.seq)

        reference = Reference(name=ref_name, length=ref_length, sequences=list(ref_contigs.values()))

        self._connection.get_session().add(reference)
        self._connection.get_session().commit()

    def update_tree(self, reference_name: str, tree: Tree, alignment_length: int) -> None:
        if alignment_length is None or alignment_length <= 0:
            raise Exception(f'Invalid alignment_length=[{alignment_length}]')

        reference = self.find_reference_genome(reference_name)
        reference.tree = tree
        reference.tree_alignment_length = alignment_length
        self._connection.get_session().commit()

    def find_reference_genome(self, name: str) -> Reference:
        try:
            return self._connection.get_session().query(Reference).filter_by(name=name).one()
        except NoResultFound as e:
            raise EntityExistsError(f'No reference genome with name=[{name}]')

    def exists_reference_genome(self, name: str) -> bool:
        return self._connection.get_session().query(Reference.id).filter_by(name=name).scalar() is not None

    def get_reference_genomes(self) -> List[Reference]:
        return self._connection.get_session().query(Reference).all()

    def count_reference_genomes(self) -> int:
        return self._connection.get_session().query(Reference).count()

    def translate_spdi(self, spdi_ids: Iterable, to: str = 'spdi_ref') -> Dict[str, str]:
        if to not in self.MUTATION_ID_TYPES:
            raise Exception(f'to={to} must be one of {self.MUTATION_ID_TYPES}')

        # Convert to query feature mutations so I can get spdi parts
        spdi_features = {QueryFeatureMutationSPDI(i) for i in spdi_ids}

        # Get set of sequence names since it's likely all (or most) of the input spdi_ids are on the same sequence
        sequence_names = {f.sequence for f in spdi_features}

        # Create map of seq name to sequences so I don't have to re-generate them for every spdi_id
        namespace = self._seq_repo_namespace
        name_sequence_map = {n: self._seq_repo_proxy.get_sequence(f'{namespace}:{n}') for n in sequence_names}

        spdi_ids_map = {}
        for f in spdi_features:
            reference_sequence_str = name_sequence_map[f.sequence]
            spdi_del_str = reference_sequence_str[f.start0:f.stop0]
            spdi_ref = f'{f.sequence}:{f.position}:{spdi_del_str}:{f.insertion}'
            spdi_ids_map[f.id] = spdi_ref

        return spdi_ids_map

    def find_references_for_sample(self, sample_name: str) -> List[Reference]:
        return self._connection.get_session().query(Reference) \
            .join(Reference.sample_nucleotide_variation) \
            .join(SampleNucleotideVariation.sample) \
            .filter(Sample.name == sample_name) \
            .all()

    def find_reference_for_sequence(self, sequence_name: str) -> Reference:
        return self._connection.get_session().query(Reference) \
            .join(Reference.sequences) \
            .filter(ReferenceSequence.sequence_name == sequence_name) \
            .one()



from __future__ import annotations

import logging
import multiprocessing as mp
import os
import tempfile
import uuid
from pathlib import Path
from typing import List, Tuple, Optional

import pandas as pd
import vcf
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

from genomics_data_index.storage.io.mutation.SnpEffDatabase import SnpEffDatabase
from genomics_data_index.storage.io.mutation.VcfSnpEffAnnotationParser import VcfSnpEffAnnotationParser
from genomics_data_index.storage.model.NucleotideMutationTranslater import NucleotideMutationTranslater
from genomics_data_index.storage.util import execute_commands

logger = logging.getLogger(__name__)


def translate_to_mutation_id(x: pd.Series) -> str:
    return NucleotideMutationTranslater.to_spdi(sequence_name=x['CHROM'],
                                                position=x['POS'],
                                                ref=x['REF'],
                                                alt=x['ALT'],
                                                convert_deletion=False)


class VariationFile:
    VCF_FRAME_COLUMNS = ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID']

    def __init__(self, file: Path):
        self._file = file

    def write(self, output: Path) -> Tuple[Path, Path]:
        if output.suffix == '.bcf':
            output_type = 'b'
        elif str(output).endswith('.vcf.gz'):
            output_type = 'z'
        else:
            raise Exception((f'Invalid file type [{output.suffix}] for output=[{output}]. '
                             'Must be one of [".bcf", ".vcf.bz"]'))

        output_index = Path(str(output) + '.csi')

        execute_commands([
            ['bcftools', 'plugin', 'fill-tags', str(self._file), '-O', output_type, '-o', str(output),
             '--', '-t', 'TYPE'],
            ['bcftools', 'index', str(output)]
        ])
        return output, output_index

    def _fix_df_columns(self, vcf_df: pd.DataFrame) -> pd.DataFrame:
        # If no data, I still want certain column names so that rest of code still works
        if len(vcf_df) == 0:
            vcf_df = pd.DataFrame(columns=['CHROM', 'POS', 'REF', 'ALT', 'INFO'])

        return vcf_df.loc[:, ['CHROM', 'POS', 'REF', 'ALT', 'INFO']]

    def _fix_alt(self, element: List[str]) -> str:
        """
        Fix up the alternative string as the pyVCF package does not return them as a string.
        :param element: The element to fix.
        :return: The fixed element.
        """
        return str(element[0])

    def _fix_ref(self, element: List[str]) -> str:
        """
        Fix up the reference string as the pyVCF package does not return them as a string.
        :param element: The element to fix.
        :return: The fixed element.
        """
        return str(element)

    def _get_type(self, vcf_df: pd.DataFrame) -> pd.Series:
        return vcf_df['INFO'].map(lambda x: x['TYPE'][0])

    def _drop_extra_columns(self, vcf_df: pd.DataFrame) -> pd.DataFrame:
        return vcf_df.drop('INFO', axis='columns')

    def read_features(self, sample_name: str, snpeff_parser: VcfSnpEffAnnotationParser) -> pd.DataFrame:
        reader = vcf.Reader(filename=str(self.file))
        df = pd.DataFrame([vars(r) for r in reader])
        out = self._fix_df_columns(df)

        out['ALT'] = out['ALT'].map(self._fix_alt)
        out['REF'] = out['REF'].map(self._fix_ref)
        out['TYPE'] = self._get_type(out)

        snpeff_headers = snpeff_parser.parse_annotation_headers(vcf_info=reader.infos)
        ann_df = snpeff_parser.parse_annotation_entries(vcf_ann_headers=snpeff_headers, vcf_df=out)
        out = out.merge(ann_df, how='left', left_index=True, right_index=True)

        out = self._drop_extra_columns(out)

        out['FILE'] = os.path.basename(self.file)
        cols = out.columns.tolist()
        out['SAMPLE'] = sample_name
        out = out.reindex(columns=['SAMPLE'] + cols)
        out = snpeff_parser.select_variant_annotations(out)
        return out.loc[:, self.VCF_FRAME_COLUMNS + snpeff_parser.ANNOTATION_COLUMNS]

    @property
    def file(self) -> Path:
        return self._file

    def annotate(self, snpeff_database: SnpEffDatabase, annotated_vcf: Path) -> VariationFile:
        return VariationFile(snpeff_database.annotate(self._file, output_vcf_file=annotated_vcf))

    def consensus(self, reference_file: Path, mask_file: Path = None, include_expression: str = 'TYPE="SNP"',
                  mask_with: str = 'N') -> List[SeqRecord]:
        with tempfile.NamedTemporaryFile() as out_f:
            command = ['bcftools', 'consensus', '--fasta-ref', str(reference_file),
                       '--mark-del', '-']
            if mask_file is not None:
                command.extend(['--mask', str(mask_file), '--mask-with', mask_with])
            command.extend([
                '--include', include_expression,
                '--output', str(out_f.name),
                str(self._file)
            ])
            execute_commands([command])
            sequences = list(SeqIO.parse(out_f.name, 'fasta'))
            return sequences

    @classmethod
    def union_all_files(cls, variant_files: List[Path], include_expression: Optional[str] = None,
                        ncores=1, batch_size=50) -> pd.DataFrame:
        if len(variant_files) == 0:
            empty_df = pd.DataFrame([], columns=['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'])
            empty_df['POS'] = empty_df['POS'].astype(int)
            return empty_df
        else:
            with tempfile.TemporaryDirectory() as tmp_dir_str:
                # First batch up files to be processed (tools cannot handle too many command-line arguments and
                # bcftools isec does not support reading list of file names from a file)
                # Plus, batch processing means I can distribute over multiple processes
                tmp_dir = Path(tmp_dir_str)
                batches = []
                tmp_file = Path(tmp_dir, str(uuid.uuid1()))
                batch = BcfToolsUnionExecutor(nthreads=1, include_expression=include_expression,
                                              out_file=tmp_file)
                batches.append(batch)
                num_in_batch = 0
                for file in variant_files:
                    batch.add_file(file)
                    num_in_batch += 1
                    # start new batch
                    if num_in_batch == batch_size:
                        tmp_file = Path(tmp_dir, str(uuid.uuid1()))
                        batch = BcfToolsUnionExecutor(nthreads=1, include_expression=include_expression,
                                                      out_file=tmp_file)
                        batches.append(batch)
                        num_in_batch = 0

                # Execute batches in a process pool
                logger.debug(f'Started processing {len(variant_files)} VCF/BCF files in batches of {batch_size}')
                with mp.Pool(ncores) as pool:
                    union_df_results = []

                    union_dfs = pool.map(cls._single_batch_union, batches)

                    for union_df in union_dfs:
                        union_df_results.append(union_df)

                    var_df = pd.concat(union_df_results).groupby('ID').agg({
                        'ID': 'first',
                        'CHROM': 'first',
                        'POS': 'first',
                        'REF': 'first',
                        'ALT': 'first',
                        'COUNT': 'sum',
                    })
                    var_df['POS'] = var_df['POS'].astype(int)

                    logger.debug(f'Finished processing {len(variant_files)} VCF/BCF files')
                    return var_df.sort_values(['CHROM', 'POS'])

    @classmethod
    def _single_batch_union(cls, union_executor: BcfToolsUnionExecutor) -> pd.DataFrame:
        return union_executor.union()


class BcfToolsUnionExecutor:

    def __init__(self, nthreads: int, include_expression: str, out_file: Path):
        self._files = []
        self._nthreads = nthreads
        self._include_expression = include_expression
        self._out_file = out_file

    def add_file(self, file: Path):
        self._files.append(file)

    def union(self) -> pd.DataFrame:
        if len(self._files) == 0:
            empty_df = pd.DataFrame([], columns=['ID', 'CHROM', 'POS', 'REF', 'ALT', 'COUNT'])
            empty_df['POS'] = empty_df['POS'].astype(int)
            return empty_df
        elif len(self._files) == 1:
            return self._union_single(self._files[0])
        else:
            return self._union_multiple(self._files)

    def _union_single(self, file: Path) -> pd.Dataframe:
        command = ['bcftools', 'query', '-f', '%CHROM\t%POS\t%REF\t%ALT\n']
        if self._include_expression is not None:
            command.extend(['-i', self._include_expression])
        command.extend(['-o', str(self._out_file), str(file)])

        execute_commands([
            command
        ])
        var_df = pd.read_csv(self._out_file, sep='\t', dtype=str,
                             names=['CHROM', 'POS', 'REF', 'ALT'])
        var_df['POS'] = var_df['POS'].astype(int)

        # Only single file so count is 1
        var_df['COUNT'] = 1

        # Added this ID so I can group by later
        if len(var_df) == 0:
            var_df['ID'] = None
        else:
            var_df['ID'] = var_df.apply(translate_to_mutation_id, axis='columns')

        return var_df.sort_values(['CHROM', 'POS'])

    def _union_multiple(self, files: List[Path]) -> pd.DataFrame:
        command = ['bcftools', 'isec']
        if self._include_expression is not None:
            command.extend(['-i', self._include_expression])
        command.extend(['-c', 'none', '-n', '+1', '--threads', f'{self._nthreads}', '-o', str(self._out_file)])
        for file in files:
            command.append(str(file))

        execute_commands([
            command
        ])

        var_df = pd.read_csv(self._out_file, sep='\t', dtype=str,
                             names=['CHROM', 'POS', 'REF', 'ALT', 'INDEXES'])
        # Count occurences of '1' character in INDEXES which represents number of samples with this mutation
        var_df['COUNT'] = var_df['INDEXES'].apply(lambda x: x.count('1'))
        var_df = var_df.drop(columns=['INDEXES'])

        var_df['POS'] = var_df['POS'].astype(int)

        # Added this ID so I can group by later
        var_df['ID'] = var_df.apply(translate_to_mutation_id, axis='columns')

        return var_df



#!/usr/bin/env python
# Prints the length of all intervals in a bed file
# Usage: cat file.bed | print-length-bed.py 

import sys

from pybedtools import BedTool

lines = ''.join(sys.stdin.readlines())
bed = BedTool(lines, from_string=True)

total = 0
for i in bed:
    total += len(i)

print(total)



import logging
import math
from os import path
from pathlib import Path
from shutil import copy
from typing import List

import pandas as pd
import yaml

from genomics_data_index.pipelines.ExecutorResults import ExecutorResults
from genomics_data_index.pipelines.PipelineExecutor import PipelineExecutor
from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile
from genomics_data_index.storage.util import execute_commands

logger = logging.getLogger(__name__)

snakemake_file = Path(path.dirname(__file__), 'snakemake', 'main', 'workflow', 'Snakefile')


class SnakemakePipelineExecutor(PipelineExecutor):

    def __init__(self, working_directory: Path = None, use_conda: bool = True,
                 include_kmer: bool = True, include_mlst: bool = True,
                 ignore_snpeff: bool = False,
                 kmer_sizes: List[int] = None, kmer_scaled: int = 1000,
                 snakemake_input_batch_size: int = 5000,
                 reads_mincov: int = 10, reads_minqual: float = 100,
                 reads_subsample: float = 1):
        super().__init__()
        if kmer_sizes is None:
            kmer_sizes = [31]

        self._working_directory = working_directory
        self._use_conda = use_conda
        self._include_kmer = include_kmer
        self._include_mlst = include_mlst
        self._ignore_snpeff = ignore_snpeff
        self._snakemake_batch_size = snakemake_input_batch_size
        self._sourmash_params = self._prepare_sourmash_params(kmer_sizes=kmer_sizes, kmer_scaled=kmer_scaled)
        self._reads_mincov = reads_mincov
        self._reads_minqual = reads_minqual
        self._reads_subsample = reads_subsample

    def _prepare_sourmash_params(self, kmer_sizes: List[int], kmer_scaled: int) -> str:
        params = ','.join([f'k={v}' for v in kmer_sizes])
        params += f',scaled={kmer_scaled}'

        return params

    def _prepare_working_directory(self, reference_file: Path,
                                   input_sample_files: pd.DataFrame) -> Path:
        config_dir = self._working_directory / 'config'
        config_dir.mkdir()

        config_file = config_dir / 'config.yaml'
        samples_file = config_dir / 'samples.tsv'

        can_use_snpeff_reference = SequenceFile(reference_file).can_use_snpeff()
        include_snpeff = False
        if can_use_snpeff_reference and not self._ignore_snpeff:
            include_snpeff = True
            logger.info('Including snpeff annotations in snakemake results')
        elif can_use_snpeff_reference:
            logger.info(f'Can use snpeff for reference file [{reference_file}] but ignore_snpeff=[{include_snpeff}]'
                        f' so will not use snpeff.')
        else:
            logger.info(f'Cannot use snpeff for reference file [{reference_file}], no snpeff annotations are included')

        logger.debug(f'Writing snakemake config [{config_file}]')
        with open(config_file, 'w') as fh:
            config = {
                'reference': str(reference_file.absolute()),
                'samples': str(samples_file.absolute()),
                'include_mlst': self._include_mlst,
                'include_kmer': self._include_kmer,
                'include_snpeff': include_snpeff,
                'sourmash_params': self._sourmash_params,
                'reads_mincov': self._reads_mincov,
                'reads_minqual': self._reads_minqual,
                'reads_subsample': self._reads_subsample,
            }
            yaml.dump(config, fh)
            logger.debug(f'Snakemake config={config}')

        logger.debug(f'Writing samples list [{samples_file}]')
        self.write_input_sample_files(input_sample_files=input_sample_files, output_file=samples_file,
                                      abolute_paths=True)

        return config_file

    def _apply_use_conda(self, command: List[str]) -> List[str]:
        if self._use_conda:
            logger.debug('Enabling --use-conda for snakemake pipeline')
            return command + ['--use-conda']
        else:
            return command

    def _get_number_batches(self, number_input_files: int) -> int:
        return int(math.ceil(number_input_files / self._snakemake_batch_size))

    def execute(self, sample_files: pd.DataFrame, reference_file: Path, ncores: int = 1) -> ExecutorResults:
        working_directory = self._working_directory

        # Preconditions
        if working_directory is None:
            raise Exception(f'working_directory is None. Please re-create {self.__class__.__name__} with a proper '
                            f'working directory.')
        self.validate_input_sample_files(sample_files)
        sample_files, sample_names_original_fixed = self.fix_sample_names(sample_files)

        number_samples = len(sample_files)
        logger.debug(f'Preparing working directory [{working_directory}] for snakemake')
        config_file = self._prepare_working_directory(reference_file=reference_file,
                                                      input_sample_files=sample_files)

        logger.debug(f'Executing snakemake on {number_samples} files with reference_file=[{reference_file}]'
                     f' using {ncores} cores in [{working_directory}]')
        snakemake_output_fofn = working_directory / 'gdi-input.fofn'
        snakemake_output_mlst = working_directory / 'mlst.tsv'

        command = ['snakemake', '--configfile', str(config_file)]
        command = self._apply_use_conda(command)
        command = command + ['-j', str(ncores), '--directory', str(working_directory),
                             '--snakefile', str(snakemake_file)]

        # I found with very large numbers of input files snakemake would slow down significantly
        # So I'm executing batches of files in snakemake if you pass too many input files.
        number_batches = self._get_number_batches(number_samples)
        if number_batches > 1:
            for batch_number in range(1, number_batches + 1):
                logger.info(f'Running snakemake batch: {batch_number}/{number_batches}')
                batch_command_gdi = command + ['--batch', f'gdi_input_fofn={batch_number}/{number_batches}']
                execute_commands([batch_command_gdi])

                # Also batch MLST results
                if self._include_mlst:
                    batch_command_mlst = command + ['--batch', f'mlst_full_table={batch_number}/{number_batches}']
                    execute_commands([batch_command_mlst])

        logger.info('Running Snakemake for rule all')
        execute_commands([command])
        logger.info('Finished running snakemake.')

        if sample_names_original_fixed is not None:
            logger.debug(f'Restoring original sample names for {len(sample_files)} samples')
            self._restore_and_write_sample_names(snakemake_output_fofn, sample_column='Sample',
                                                 samples_original_fixed=sample_names_original_fixed)
        else:
            logger.debug('Original sample names were unmodified so no need to restore them')

        logger.debug(f'Output file [{snakemake_output_fofn}]. '
                     f'MLST file [{snakemake_output_mlst}]')

        return ExecutorResults({'gdi-fofn': snakemake_output_fofn, 'mlst': snakemake_output_mlst})

    def _restore_and_write_sample_names(self, file: Path, sample_column: str,
                                        samples_original_fixed: pd.DataFrame) -> None:
        backup_file = Path(str(file) + '.bak')
        copy(file, backup_file)

        file_df = pd.read_csv(file, sep='\t')
        file_df = self.restore_sample_names(file_df, sample_column=sample_column,
                                            samples_original_fixed=samples_original_fixed)
        file_df.to_csv(file, sep='\t', index=False)
